# coding: utf-8

"""
    Managed Ray API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from openapi_client.configuration import Configuration


class UserMigrationResult(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'user_id': 'str',
        'user_email': 'str',
        'identity_id': 'str',
        'cloud_permissions_to_remove': 'list[CloudPermissionRemoval]',
        'project_permissions_to_remove': 'list[ProjectPermissionRemoval]',
        'readonly_overrides_to_remove': 'list[str]',
        'org_permission_changes': 'list[OrgPermissionChange]',
        'cloud_permissions_deleted': 'int',
        'project_permissions_deleted': 'int',
        'readonly_permissions_deleted': 'int',
        'org_owner_demoted': 'bool'
    }

    attribute_map = {
        'user_id': 'user_id',
        'user_email': 'user_email',
        'identity_id': 'identity_id',
        'cloud_permissions_to_remove': 'cloud_permissions_to_remove',
        'project_permissions_to_remove': 'project_permissions_to_remove',
        'readonly_overrides_to_remove': 'readonly_overrides_to_remove',
        'org_permission_changes': 'org_permission_changes',
        'cloud_permissions_deleted': 'cloud_permissions_deleted',
        'project_permissions_deleted': 'project_permissions_deleted',
        'readonly_permissions_deleted': 'readonly_permissions_deleted',
        'org_owner_demoted': 'org_owner_demoted'
    }

    def __init__(self, user_id=None, user_email=None, identity_id=None, cloud_permissions_to_remove=None, project_permissions_to_remove=None, readonly_overrides_to_remove=None, org_permission_changes=None, cloud_permissions_deleted=0, project_permissions_deleted=0, readonly_permissions_deleted=0, org_owner_demoted=None, local_vars_configuration=None):  # noqa: E501
        """UserMigrationResult - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._user_id = None
        self._user_email = None
        self._identity_id = None
        self._cloud_permissions_to_remove = None
        self._project_permissions_to_remove = None
        self._readonly_overrides_to_remove = None
        self._org_permission_changes = None
        self._cloud_permissions_deleted = None
        self._project_permissions_deleted = None
        self._readonly_permissions_deleted = None
        self._org_owner_demoted = None
        self.discriminator = None

        self.user_id = user_id
        self.user_email = user_email
        self.identity_id = identity_id
        if cloud_permissions_to_remove is not None:
            self.cloud_permissions_to_remove = cloud_permissions_to_remove
        if project_permissions_to_remove is not None:
            self.project_permissions_to_remove = project_permissions_to_remove
        if readonly_overrides_to_remove is not None:
            self.readonly_overrides_to_remove = readonly_overrides_to_remove
        if org_permission_changes is not None:
            self.org_permission_changes = org_permission_changes
        if cloud_permissions_deleted is not None:
            self.cloud_permissions_deleted = cloud_permissions_deleted
        if project_permissions_deleted is not None:
            self.project_permissions_deleted = project_permissions_deleted
        if readonly_permissions_deleted is not None:
            self.readonly_permissions_deleted = readonly_permissions_deleted
        self.org_owner_demoted = org_owner_demoted

    @property
    def user_id(self):
        """Gets the user_id of this UserMigrationResult.  # noqa: E501


        :return: The user_id of this UserMigrationResult.  # noqa: E501
        :rtype: str
        """
        return self._user_id

    @user_id.setter
    def user_id(self, user_id):
        """Sets the user_id of this UserMigrationResult.


        :param user_id: The user_id of this UserMigrationResult.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and user_id is None:  # noqa: E501
            raise ValueError("Invalid value for `user_id`, must not be `None`")  # noqa: E501

        self._user_id = user_id

    @property
    def user_email(self):
        """Gets the user_email of this UserMigrationResult.  # noqa: E501


        :return: The user_email of this UserMigrationResult.  # noqa: E501
        :rtype: str
        """
        return self._user_email

    @user_email.setter
    def user_email(self, user_email):
        """Sets the user_email of this UserMigrationResult.


        :param user_email: The user_email of this UserMigrationResult.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and user_email is None:  # noqa: E501
            raise ValueError("Invalid value for `user_email`, must not be `None`")  # noqa: E501

        self._user_email = user_email

    @property
    def identity_id(self):
        """Gets the identity_id of this UserMigrationResult.  # noqa: E501


        :return: The identity_id of this UserMigrationResult.  # noqa: E501
        :rtype: str
        """
        return self._identity_id

    @identity_id.setter
    def identity_id(self, identity_id):
        """Sets the identity_id of this UserMigrationResult.


        :param identity_id: The identity_id of this UserMigrationResult.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and identity_id is None:  # noqa: E501
            raise ValueError("Invalid value for `identity_id`, must not be `None`")  # noqa: E501

        self._identity_id = identity_id

    @property
    def cloud_permissions_to_remove(self):
        """Gets the cloud_permissions_to_remove of this UserMigrationResult.  # noqa: E501


        :return: The cloud_permissions_to_remove of this UserMigrationResult.  # noqa: E501
        :rtype: list[CloudPermissionRemoval]
        """
        return self._cloud_permissions_to_remove

    @cloud_permissions_to_remove.setter
    def cloud_permissions_to_remove(self, cloud_permissions_to_remove):
        """Sets the cloud_permissions_to_remove of this UserMigrationResult.


        :param cloud_permissions_to_remove: The cloud_permissions_to_remove of this UserMigrationResult.  # noqa: E501
        :type: list[CloudPermissionRemoval]
        """

        self._cloud_permissions_to_remove = cloud_permissions_to_remove

    @property
    def project_permissions_to_remove(self):
        """Gets the project_permissions_to_remove of this UserMigrationResult.  # noqa: E501


        :return: The project_permissions_to_remove of this UserMigrationResult.  # noqa: E501
        :rtype: list[ProjectPermissionRemoval]
        """
        return self._project_permissions_to_remove

    @project_permissions_to_remove.setter
    def project_permissions_to_remove(self, project_permissions_to_remove):
        """Sets the project_permissions_to_remove of this UserMigrationResult.


        :param project_permissions_to_remove: The project_permissions_to_remove of this UserMigrationResult.  # noqa: E501
        :type: list[ProjectPermissionRemoval]
        """

        self._project_permissions_to_remove = project_permissions_to_remove

    @property
    def readonly_overrides_to_remove(self):
        """Gets the readonly_overrides_to_remove of this UserMigrationResult.  # noqa: E501


        :return: The readonly_overrides_to_remove of this UserMigrationResult.  # noqa: E501
        :rtype: list[str]
        """
        return self._readonly_overrides_to_remove

    @readonly_overrides_to_remove.setter
    def readonly_overrides_to_remove(self, readonly_overrides_to_remove):
        """Sets the readonly_overrides_to_remove of this UserMigrationResult.


        :param readonly_overrides_to_remove: The readonly_overrides_to_remove of this UserMigrationResult.  # noqa: E501
        :type: list[str]
        """

        self._readonly_overrides_to_remove = readonly_overrides_to_remove

    @property
    def org_permission_changes(self):
        """Gets the org_permission_changes of this UserMigrationResult.  # noqa: E501


        :return: The org_permission_changes of this UserMigrationResult.  # noqa: E501
        :rtype: list[OrgPermissionChange]
        """
        return self._org_permission_changes

    @org_permission_changes.setter
    def org_permission_changes(self, org_permission_changes):
        """Sets the org_permission_changes of this UserMigrationResult.


        :param org_permission_changes: The org_permission_changes of this UserMigrationResult.  # noqa: E501
        :type: list[OrgPermissionChange]
        """

        self._org_permission_changes = org_permission_changes

    @property
    def cloud_permissions_deleted(self):
        """Gets the cloud_permissions_deleted of this UserMigrationResult.  # noqa: E501


        :return: The cloud_permissions_deleted of this UserMigrationResult.  # noqa: E501
        :rtype: int
        """
        return self._cloud_permissions_deleted

    @cloud_permissions_deleted.setter
    def cloud_permissions_deleted(self, cloud_permissions_deleted):
        """Sets the cloud_permissions_deleted of this UserMigrationResult.


        :param cloud_permissions_deleted: The cloud_permissions_deleted of this UserMigrationResult.  # noqa: E501
        :type: int
        """

        self._cloud_permissions_deleted = cloud_permissions_deleted

    @property
    def project_permissions_deleted(self):
        """Gets the project_permissions_deleted of this UserMigrationResult.  # noqa: E501


        :return: The project_permissions_deleted of this UserMigrationResult.  # noqa: E501
        :rtype: int
        """
        return self._project_permissions_deleted

    @project_permissions_deleted.setter
    def project_permissions_deleted(self, project_permissions_deleted):
        """Sets the project_permissions_deleted of this UserMigrationResult.


        :param project_permissions_deleted: The project_permissions_deleted of this UserMigrationResult.  # noqa: E501
        :type: int
        """

        self._project_permissions_deleted = project_permissions_deleted

    @property
    def readonly_permissions_deleted(self):
        """Gets the readonly_permissions_deleted of this UserMigrationResult.  # noqa: E501


        :return: The readonly_permissions_deleted of this UserMigrationResult.  # noqa: E501
        :rtype: int
        """
        return self._readonly_permissions_deleted

    @readonly_permissions_deleted.setter
    def readonly_permissions_deleted(self, readonly_permissions_deleted):
        """Sets the readonly_permissions_deleted of this UserMigrationResult.


        :param readonly_permissions_deleted: The readonly_permissions_deleted of this UserMigrationResult.  # noqa: E501
        :type: int
        """

        self._readonly_permissions_deleted = readonly_permissions_deleted

    @property
    def org_owner_demoted(self):
        """Gets the org_owner_demoted of this UserMigrationResult.  # noqa: E501


        :return: The org_owner_demoted of this UserMigrationResult.  # noqa: E501
        :rtype: bool
        """
        return self._org_owner_demoted

    @org_owner_demoted.setter
    def org_owner_demoted(self, org_owner_demoted):
        """Sets the org_owner_demoted of this UserMigrationResult.


        :param org_owner_demoted: The org_owner_demoted of this UserMigrationResult.  # noqa: E501
        :type: bool
        """
        if self.local_vars_configuration.client_side_validation and org_owner_demoted is None:  # noqa: E501
            raise ValueError("Invalid value for `org_owner_demoted`, must not be `None`")  # noqa: E501

        self._org_owner_demoted = org_owner_demoted

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, UserMigrationResult):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, UserMigrationResult):
            return True

        return self.to_dict() != other.to_dict()



import json
from json import dumps as json_dumps
from typing import Dict, Optional

import click
from rich.console import Console
from rich.table import Table

import anyscale
from anyscale._private.anyscale_client import AnyscaleClient
from anyscale.cli_logger import BlockLogger
from anyscale.commands import command_examples
from anyscale.commands.list_util import (
    display_list,
    MAX_PAGE_SIZE,
    NON_INTERACTIVE_DEFAULT_MAX_ITEMS,
    validate_page_size,
)
from anyscale.commands.util import AnyscaleCommand
from anyscale.user.models import AdminCreateUser, AdminCreateUsers, User
from anyscale.util import AnyscaleJSONEncoder, get_endpoint, validate_non_negative_arg


log = BlockLogger()  # CLI Logger


_COLLABORATOR_TYPE_CHOICES = (
    "all_accounts",
    "only_service_accounts",
    "only_user_accounts",
)


def _create_user_list_table(show_header: bool) -> Table:
    table = Table(show_header=show_header, expand=True)
    table.add_column("EMAIL", no_wrap=False, overflow="fold", ratio=3, min_width=20)
    table.add_column("NAME", no_wrap=False, overflow="fold", ratio=2, min_width=12)
    table.add_column("ID", no_wrap=False, overflow="fold", ratio=2, min_width=12)
    table.add_column("ROLE", no_wrap=False, overflow="fold", ratio=1, min_width=10)
    table.add_column(
        "CREATED AT", no_wrap=False, overflow="fold", ratio=1, min_width=14
    )
    return table


def _format_user_output_data(user: User) -> Dict[str, str]:
    return {
        "email": user.email,
        "name": user.name,
        "id": user.user_id or "",
        "role": user.permission_level,
        "created_at": user.created_at.strftime("%Y-%m-%d %H:%M"),
    }


@click.group("user", help="Manage users.")
def user_cli() -> None:
    pass


@user_cli.command(
    name="batch-create",
    cls=AnyscaleCommand,
    example=command_examples.USER_BATCH_CREATE_EXAMPLE,
)
@click.option(
    "--users-file",
    "-f",
    required=True,
    type=str,
    help="Path to a YAML file that contains the information for user accounts to be created.",
)
def admin_batch_create(users_file: str,) -> None:
    """
    Batch create, as an admin, users without email verification.
    """
    log.info("Creating users...")

    create_users = AdminCreateUsers.from_yaml(users_file)

    try:
        created_users = anyscale.user.admin_batch_create(
            admin_create_users=[
                AdminCreateUser(**create_user)
                for create_user in create_users.create_users
            ]
        )
    except ValueError as e:
        log.error(f"Error creating users: {e}")
        return

    log.info(f"{len(created_users)} users created.")


@user_cli.command(
    name="list",
    cls=AnyscaleCommand,
    example=command_examples.USER_LIST_EXAMPLE,
    help="List users within your organization.",
)
@click.option("--email", type=str, help="Filter users by email.")
@click.option("--name", type=str, help="Filter users by display name.")
@click.option(
    "--collaborator-type",
    type=click.Choice(_COLLABORATOR_TYPE_CHOICES, case_sensitive=False),
    help="Filter users by collaborator type.",
)
@click.option(
    "--service-account",
    "service_account",
    flag_value=True,
    help="Only show service accounts.",
)
@click.option(
    "--user-account",
    "service_account",
    flag_value=False,
    help="Only show user accounts (non-service accounts).",
)
@click.option(
    "--max-items",
    type=int,
    default=None,
    callback=validate_non_negative_arg,
    help="Maximum number of users to display when running non-interactively.",
)
@click.option(
    "--page-size",
    type=int,
    default=MAX_PAGE_SIZE,
    show_default=True,
    callback=validate_page_size,
    help=f"Items per page (max {MAX_PAGE_SIZE}).",
)
@click.option(
    "--json/--no-json",
    "json_output",
    default=False,
    help="Render output as JSON instead of a table.",
)
@click.option(
    "--interactive/--no-interactive",
    default=True,
    help="Enable or disable interactive pagination.",
)
def list_users(  # noqa: A001, PLR0913
    email: Optional[str],
    name: Optional[str],
    collaborator_type: Optional[str],
    service_account: Optional[bool],
    max_items: Optional[int],
    page_size: int,
    json_output: bool,
    interactive: bool,
) -> None:
    """List users within the organization."""
    if max_items is not None and interactive:
        raise click.UsageError("--max-items only allowed with --no-interactive")

    effective_max = max_items
    stderr = Console(stderr=True)
    if not interactive and effective_max is None:
        stderr.print(
            f"Defaulting to {NON_INTERACTIVE_DEFAULT_MAX_ITEMS} items in batch mode; "
            "use --max-items to override."
        )
        effective_max = NON_INTERACTIVE_DEFAULT_MAX_ITEMS

    console = Console()

    stderr.print("[bold]Listing users with:[/]")
    stderr.print(f"• email             = {email or '<any>'}")
    stderr.print(f"• name              = {name or '<any>'}")
    stderr.print(f"• collaborator type = {collaborator_type or '<any>'}")
    stderr.print(
        f"• service account   = "
        f"{service_account if service_account is not None else '<any>'}"
    )
    stderr.print(f"• mode              = {'interactive' if interactive else 'batch'}")
    stderr.print(f"• per-page limit    = {page_size}")
    stderr.print(f"• max-items total   = {effective_max or 'all'}")
    stderr.print(f"\nView your Users in the UI at {get_endpoint('/collaborators')}\n")

    collaborator_type_value = collaborator_type.lower() if collaborator_type else None

    formatter = (
        (lambda user: user.to_dict()) if json_output else _format_user_output_data
    )

    try:
        iterator = anyscale.user.list(
            email=email,
            name=name,
            collaborator_type=collaborator_type_value,
            is_service_account=service_account,
            max_items=None if interactive else effective_max,
            page_size=page_size,
        )

        total = display_list(
            iterator=iter(iterator),
            item_formatter=formatter,
            table_creator=_create_user_list_table,
            json_output=json_output,
            page_size=page_size,
            interactive=interactive,
            max_items=effective_max,
            console=console,
        )

        if not json_output:
            if total > 0:
                stderr.print(f"\nFetched {total} users.")
            else:
                stderr.print("\nNo users found.")
    except ValueError as exc:
        raise click.ClickException(str(exc))
    except Exception as exc:  # noqa: BLE001
        raise click.ClickException(f"Failed to list users: {exc}")


@user_cli.command(
    name="get",
    cls=AnyscaleCommand,
    example=command_examples.USER_GET_EXAMPLE,
    help="Get details for a single user in your organization.",
)
@click.option("--email", type=str, help="Email address of the user.")
@click.option("--name", type=str, help="Display name of the user.")
@click.option(
    "--collaborator-type",
    type=click.Choice(_COLLABORATOR_TYPE_CHOICES, case_sensitive=False),
    help="Optional collaborator type constraint.",
)
@click.option(
    "--service-account",
    "service_account",
    flag_value=True,
    help="Restrict to service accounts.",
)
@click.option(
    "--user-account",
    "service_account",
    flag_value=False,
    help="Restrict to individual user accounts.",
)
@click.option(
    "--json/--no-json",
    "json_output",
    default=False,
    help="Output JSON instead of YAML.",
)
def get_user(  # noqa: A001
    email: Optional[str],
    name: Optional[str],
    collaborator_type: Optional[str],
    service_account: Optional[bool],
    json_output: bool,
) -> None:
    """Retrieve details for a single user by email or ID."""
    if email is None and name is None:
        raise click.UsageError("Provide --email or --name.")

    collaborator_type_value = collaborator_type.lower() if collaborator_type else None

    console = Console()
    stderr = Console(stderr=True)

    stderr.print("[bold]Fetching user with:[/]")
    stderr.print(f"• email             = {email or '<not specified>'}")
    stderr.print(f"• name              = {name or '<not specified>'}")
    stderr.print(f"• collaborator type = {collaborator_type or '<any>'}")
    stderr.print(
        f"• service account   = "
        f"{service_account if service_account is not None else '<any>'}"
    )
    stderr.print(f"\nView your Users in the UI at {get_endpoint('/collaborators')}\n")

    try:
        user = anyscale.user.get(
            email=email,
            name=name,
            collaborator_type=collaborator_type_value,
            is_service_account=service_account,
        )
    except ValueError as exc:
        raise click.ClickException(str(exc))
    except Exception as exc:  # noqa: BLE001
        raise click.ClickException(f"Failed to get user: {exc}")

    if json_output:
        console.print_json(
            json=json_dumps(user.to_dict(), indent=2, cls=AnyscaleJSONEncoder)
        )
        return

    table = _create_user_list_table(show_header=True)
    formatted_data = _format_user_output_data(user)
    table.add_row(
        formatted_data["email"],
        formatted_data["name"],
        formatted_data["id"],
        formatted_data["role"],
        formatted_data["created_at"],
    )
    console.print(table)
    stderr.print("\nFetched 1 user.")


@user_cli.command(
    name="list-permissions",
    cls=AnyscaleCommand,
    example=command_examples.USER_LIST_PERMISSIONS_EXAMPLE,
)
@click.option(
    "--user-id",
    type=str,
    default=None,
    help="Filter to a specific user ID. If not provided, lists permissions for all users.",
)
@click.option(
    "--output",
    "-o",
    type=click.Path(),
    default=None,
    help="Write JSON output to a file instead of stdout.",
)
def list_permissions(user_id: Optional[str], output: Optional[str]) -> None:
    """
    List users and their effective cloud/project permissions across the organization.

    Shows each user's access to clouds and projects, combining both individually
    granted permissions and permissions inherited from user groups. Also lists
    the current organization owners.

    Output is JSON. Use --output to save to a file.
    """
    client = AnyscaleClient()
    try:
        log.info("Listing users and their effective permissions...")
        response = client.list_scim_user_permissions(user_id=user_id)
        result = response.get("result", response)

        # Derive per-user org role from org_owners:
        json_output = json.dumps(result, indent=2, sort_keys=True)

        if output:
            with open(output, "w") as f:
                f.write(json_output)
            log.info(f"Results written to {output}")
        else:
            print(json_output)
    except click.ClickException:
        raise
    except Exception as e:  # noqa: BLE001
        log.error(f"Failed to list user permissions: {e}")
        raise click.ClickException(str(e))



# coding: utf-8

"""
    Managed Ray API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from openapi_client.configuration import Configuration


class GPUUsage(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'gpu_index': 'int',
        'gpu_device_name': 'str',
        'gpu_usage': 'float',
        'gram_usage': 'int',
        'gram_free': 'int'
    }

    attribute_map = {
        'gpu_index': 'gpu_index',
        'gpu_device_name': 'gpu_device_name',
        'gpu_usage': 'gpu_usage',
        'gram_usage': 'gram_usage',
        'gram_free': 'gram_free'
    }

    def __init__(self, gpu_index=None, gpu_device_name=None, gpu_usage=None, gram_usage=None, gram_free=None, local_vars_configuration=None):  # noqa: E501
        """GPUUsage - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._gpu_index = None
        self._gpu_device_name = None
        self._gpu_usage = None
        self._gram_usage = None
        self._gram_free = None
        self.discriminator = None

        self.gpu_index = gpu_index
        self.gpu_device_name = gpu_device_name
        if gpu_usage is not None:
            self.gpu_usage = gpu_usage
        if gram_usage is not None:
            self.gram_usage = gram_usage
        if gram_free is not None:
            self.gram_free = gram_free

    @property
    def gpu_index(self):
        """Gets the gpu_index of this GPUUsage.  # noqa: E501

        The index of the GPU.  # noqa: E501

        :return: The gpu_index of this GPUUsage.  # noqa: E501
        :rtype: int
        """
        return self._gpu_index

    @gpu_index.setter
    def gpu_index(self, gpu_index):
        """Sets the gpu_index of this GPUUsage.

        The index of the GPU.  # noqa: E501

        :param gpu_index: The gpu_index of this GPUUsage.  # noqa: E501
        :type: int
        """
        if self.local_vars_configuration.client_side_validation and gpu_index is None:  # noqa: E501
            raise ValueError("Invalid value for `gpu_index`, must not be `None`")  # noqa: E501

        self._gpu_index = gpu_index

    @property
    def gpu_device_name(self):
        """Gets the gpu_device_name of this GPUUsage.  # noqa: E501

        The device name of the GPU.  # noqa: E501

        :return: The gpu_device_name of this GPUUsage.  # noqa: E501
        :rtype: str
        """
        return self._gpu_device_name

    @gpu_device_name.setter
    def gpu_device_name(self, gpu_device_name):
        """Sets the gpu_device_name of this GPUUsage.

        The device name of the GPU.  # noqa: E501

        :param gpu_device_name: The gpu_device_name of this GPUUsage.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and gpu_device_name is None:  # noqa: E501
            raise ValueError("Invalid value for `gpu_device_name`, must not be `None`")  # noqa: E501

        self._gpu_device_name = gpu_device_name

    @property
    def gpu_usage(self):
        """Gets the gpu_usage of this GPUUsage.  # noqa: E501

        The GPU usage in percentage of this GPU.  # noqa: E501

        :return: The gpu_usage of this GPUUsage.  # noqa: E501
        :rtype: float
        """
        return self._gpu_usage

    @gpu_usage.setter
    def gpu_usage(self, gpu_usage):
        """Sets the gpu_usage of this GPUUsage.

        The GPU usage in percentage of this GPU.  # noqa: E501

        :param gpu_usage: The gpu_usage of this GPUUsage.  # noqa: E501
        :type: float
        """

        self._gpu_usage = gpu_usage

    @property
    def gram_usage(self):
        """Gets the gram_usage of this GPUUsage.  # noqa: E501

        The GRAM usage in bytes of this GPU.  # noqa: E501

        :return: The gram_usage of this GPUUsage.  # noqa: E501
        :rtype: int
        """
        return self._gram_usage

    @gram_usage.setter
    def gram_usage(self, gram_usage):
        """Sets the gram_usage of this GPUUsage.

        The GRAM usage in bytes of this GPU.  # noqa: E501

        :param gram_usage: The gram_usage of this GPUUsage.  # noqa: E501
        :type: int
        """

        self._gram_usage = gram_usage

    @property
    def gram_free(self):
        """Gets the gram_free of this GPUUsage.  # noqa: E501

        The GRAM free in bytes of this GPU.  # noqa: E501

        :return: The gram_free of this GPUUsage.  # noqa: E501
        :rtype: int
        """
        return self._gram_free

    @gram_free.setter
    def gram_free(self, gram_free):
        """Sets the gram_free of this GPUUsage.

        The GRAM free in bytes of this GPU.  # noqa: E501

        :param gram_free: The gram_free of this GPUUsage.  # noqa: E501
        :type: int
        """

        self._gram_free = gram_free

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, GPUUsage):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, GPUUsage):
            return True

        return self.to_dict() != other.to_dict()



# coding: utf-8

"""
    Managed Ray API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from openapi_client.configuration import Configuration


class TrainRun(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'id': 'str',
        'name': 'str',
        'job_id': 'str',
        'controller_actor_id': 'str',
        'status': 'RunStatus',
        'status_detail': 'str',
        'start_time_ms': 'int',
        'end_time_ms': 'int',
        'attempts': 'list[TrainRunAttempt]',
        'total_attempts': 'int',
        'job_details': 'JobDetails',
        'session_name': 'str',
        'controller_log_id': 'str',
        'train_run_panels': 'list[DashboardPanel]',
        'train_worker_panels': 'list[DashboardPanel]'
    }

    attribute_map = {
        'id': 'id',
        'name': 'name',
        'job_id': 'job_id',
        'controller_actor_id': 'controller_actor_id',
        'status': 'status',
        'status_detail': 'status_detail',
        'start_time_ms': 'start_time_ms',
        'end_time_ms': 'end_time_ms',
        'attempts': 'attempts',
        'total_attempts': 'total_attempts',
        'job_details': 'job_details',
        'session_name': 'session_name',
        'controller_log_id': 'controller_log_id',
        'train_run_panels': 'train_run_panels',
        'train_worker_panels': 'train_worker_panels'
    }

    def __init__(self, id=None, name=None, job_id=None, controller_actor_id=None, status=None, status_detail=None, start_time_ms=None, end_time_ms=None, attempts=None, total_attempts=None, job_details=None, session_name=None, controller_log_id=None, train_run_panels=None, train_worker_panels=None, local_vars_configuration=None):  # noqa: E501
        """TrainRun - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._id = None
        self._name = None
        self._job_id = None
        self._controller_actor_id = None
        self._status = None
        self._status_detail = None
        self._start_time_ms = None
        self._end_time_ms = None
        self._attempts = None
        self._total_attempts = None
        self._job_details = None
        self._session_name = None
        self._controller_log_id = None
        self._train_run_panels = None
        self._train_worker_panels = None
        self.discriminator = None

        self.id = id
        self.name = name
        self.job_id = job_id
        if controller_actor_id is not None:
            self.controller_actor_id = controller_actor_id
        self.status = status
        if status_detail is not None:
            self.status_detail = status_detail
        self.start_time_ms = start_time_ms
        if end_time_ms is not None:
            self.end_time_ms = end_time_ms
        self.attempts = attempts
        self.total_attempts = total_attempts
        self.job_details = job_details
        if session_name is not None:
            self.session_name = session_name
        if controller_log_id is not None:
            self.controller_log_id = controller_log_id
        if train_run_panels is not None:
            self.train_run_panels = train_run_panels
        if train_worker_panels is not None:
            self.train_worker_panels = train_worker_panels

    @property
    def id(self):
        """Gets the id of this TrainRun.  # noqa: E501


        :return: The id of this TrainRun.  # noqa: E501
        :rtype: str
        """
        return self._id

    @id.setter
    def id(self, id):
        """Sets the id of this TrainRun.


        :param id: The id of this TrainRun.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and id is None:  # noqa: E501
            raise ValueError("Invalid value for `id`, must not be `None`")  # noqa: E501

        self._id = id

    @property
    def name(self):
        """Gets the name of this TrainRun.  # noqa: E501


        :return: The name of this TrainRun.  # noqa: E501
        :rtype: str
        """
        return self._name

    @name.setter
    def name(self, name):
        """Sets the name of this TrainRun.


        :param name: The name of this TrainRun.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and name is None:  # noqa: E501
            raise ValueError("Invalid value for `name`, must not be `None`")  # noqa: E501

        self._name = name

    @property
    def job_id(self):
        """Gets the job_id of this TrainRun.  # noqa: E501


        :return: The job_id of this TrainRun.  # noqa: E501
        :rtype: str
        """
        return self._job_id

    @job_id.setter
    def job_id(self, job_id):
        """Sets the job_id of this TrainRun.


        :param job_id: The job_id of this TrainRun.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and job_id is None:  # noqa: E501
            raise ValueError("Invalid value for `job_id`, must not be `None`")  # noqa: E501

        self._job_id = job_id

    @property
    def controller_actor_id(self):
        """Gets the controller_actor_id of this TrainRun.  # noqa: E501


        :return: The controller_actor_id of this TrainRun.  # noqa: E501
        :rtype: str
        """
        return self._controller_actor_id

    @controller_actor_id.setter
    def controller_actor_id(self, controller_actor_id):
        """Sets the controller_actor_id of this TrainRun.


        :param controller_actor_id: The controller_actor_id of this TrainRun.  # noqa: E501
        :type: str
        """

        self._controller_actor_id = controller_actor_id

    @property
    def status(self):
        """Gets the status of this TrainRun.  # noqa: E501


        :return: The status of this TrainRun.  # noqa: E501
        :rtype: RunStatus
        """
        return self._status

    @status.setter
    def status(self, status):
        """Sets the status of this TrainRun.


        :param status: The status of this TrainRun.  # noqa: E501
        :type: RunStatus
        """
        if self.local_vars_configuration.client_side_validation and status is None:  # noqa: E501
            raise ValueError("Invalid value for `status`, must not be `None`")  # noqa: E501

        self._status = status

    @property
    def status_detail(self):
        """Gets the status_detail of this TrainRun.  # noqa: E501


        :return: The status_detail of this TrainRun.  # noqa: E501
        :rtype: str
        """
        return self._status_detail

    @status_detail.setter
    def status_detail(self, status_detail):
        """Sets the status_detail of this TrainRun.


        :param status_detail: The status_detail of this TrainRun.  # noqa: E501
        :type: str
        """

        self._status_detail = status_detail

    @property
    def start_time_ms(self):
        """Gets the start_time_ms of this TrainRun.  # noqa: E501


        :return: The start_time_ms of this TrainRun.  # noqa: E501
        :rtype: int
        """
        return self._start_time_ms

    @start_time_ms.setter
    def start_time_ms(self, start_time_ms):
        """Sets the start_time_ms of this TrainRun.


        :param start_time_ms: The start_time_ms of this TrainRun.  # noqa: E501
        :type: int
        """
        if self.local_vars_configuration.client_side_validation and start_time_ms is None:  # noqa: E501
            raise ValueError("Invalid value for `start_time_ms`, must not be `None`")  # noqa: E501

        self._start_time_ms = start_time_ms

    @property
    def end_time_ms(self):
        """Gets the end_time_ms of this TrainRun.  # noqa: E501


        :return: The end_time_ms of this TrainRun.  # noqa: E501
        :rtype: int
        """
        return self._end_time_ms

    @end_time_ms.setter
    def end_time_ms(self, end_time_ms):
        """Sets the end_time_ms of this TrainRun.


        :param end_time_ms: The end_time_ms of this TrainRun.  # noqa: E501
        :type: int
        """

        self._end_time_ms = end_time_ms

    @property
    def attempts(self):
        """Gets the attempts of this TrainRun.  # noqa: E501


        :return: The attempts of this TrainRun.  # noqa: E501
        :rtype: list[TrainRunAttempt]
        """
        return self._attempts

    @attempts.setter
    def attempts(self, attempts):
        """Sets the attempts of this TrainRun.


        :param attempts: The attempts of this TrainRun.  # noqa: E501
        :type: list[TrainRunAttempt]
        """
        if self.local_vars_configuration.client_side_validation and attempts is None:  # noqa: E501
            raise ValueError("Invalid value for `attempts`, must not be `None`")  # noqa: E501

        self._attempts = attempts

    @property
    def total_attempts(self):
        """Gets the total_attempts of this TrainRun.  # noqa: E501


        :return: The total_attempts of this TrainRun.  # noqa: E501
        :rtype: int
        """
        return self._total_attempts

    @total_attempts.setter
    def total_attempts(self, total_attempts):
        """Sets the total_attempts of this TrainRun.


        :param total_attempts: The total_attempts of this TrainRun.  # noqa: E501
        :type: int
        """
        if self.local_vars_configuration.client_side_validation and total_attempts is None:  # noqa: E501
            raise ValueError("Invalid value for `total_attempts`, must not be `None`")  # noqa: E501

        self._total_attempts = total_attempts

    @property
    def job_details(self):
        """Gets the job_details of this TrainRun.  # noqa: E501


        :return: The job_details of this TrainRun.  # noqa: E501
        :rtype: JobDetails
        """
        return self._job_details

    @job_details.setter
    def job_details(self, job_details):
        """Sets the job_details of this TrainRun.


        :param job_details: The job_details of this TrainRun.  # noqa: E501
        :type: JobDetails
        """
        if self.local_vars_configuration.client_side_validation and job_details is None:  # noqa: E501
            raise ValueError("Invalid value for `job_details`, must not be `None`")  # noqa: E501

        self._job_details = job_details

    @property
    def session_name(self):
        """Gets the session_name of this TrainRun.  # noqa: E501


        :return: The session_name of this TrainRun.  # noqa: E501
        :rtype: str
        """
        return self._session_name

    @session_name.setter
    def session_name(self, session_name):
        """Sets the session_name of this TrainRun.


        :param session_name: The session_name of this TrainRun.  # noqa: E501
        :type: str
        """

        self._session_name = session_name

    @property
    def controller_log_id(self):
        """Gets the controller_log_id of this TrainRun.  # noqa: E501


        :return: The controller_log_id of this TrainRun.  # noqa: E501
        :rtype: str
        """
        return self._controller_log_id

    @controller_log_id.setter
    def controller_log_id(self, controller_log_id):
        """Sets the controller_log_id of this TrainRun.


        :param controller_log_id: The controller_log_id of this TrainRun.  # noqa: E501
        :type: str
        """

        self._controller_log_id = controller_log_id

    @property
    def train_run_panels(self):
        """Gets the train_run_panels of this TrainRun.  # noqa: E501


        :return: The train_run_panels of this TrainRun.  # noqa: E501
        :rtype: list[DashboardPanel]
        """
        return self._train_run_panels

    @train_run_panels.setter
    def train_run_panels(self, train_run_panels):
        """Sets the train_run_panels of this TrainRun.


        :param train_run_panels: The train_run_panels of this TrainRun.  # noqa: E501
        :type: list[DashboardPanel]
        """

        self._train_run_panels = train_run_panels

    @property
    def train_worker_panels(self):
        """Gets the train_worker_panels of this TrainRun.  # noqa: E501


        :return: The train_worker_panels of this TrainRun.  # noqa: E501
        :rtype: list[DashboardPanel]
        """
        return self._train_worker_panels

    @train_worker_panels.setter
    def train_worker_panels(self, train_worker_panels):
        """Sets the train_worker_panels of this TrainRun.


        :param train_worker_panels: The train_worker_panels of this TrainRun.  # noqa: E501
        :type: list[DashboardPanel]
        """

        self._train_worker_panels = train_worker_panels

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, TrainRun):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, TrainRun):
            return True

        return self.to_dict() != other.to_dict()



import json
from typing import Optional

import click
from rich import print as rprint
import tabulate

import anyscale
from anyscale._private.anyscale_client import AnyscaleClient
from anyscale.cli_logger import BlockLogger
from anyscale.commands import command_examples
from anyscale.commands.util import AnyscaleCommand


log = BlockLogger()


@click.group("user-group", help="Manage user groups.")
def user_group_cli() -> None:
    pass


@user_group_cli.command(
    name="list",
    cls=AnyscaleCommand,
    example=command_examples.USER_GROUP_LIST_EXAMPLE,
    is_beta=True,
)
@click.option(
    "--max-items",
    default=50,
    type=int,
    help="Maximum number of user groups to return.",
)
def list_user_groups(max_items: int) -> None:
    """
    List user groups in the organization.
    """
    try:
        user_groups = anyscale.user_group.list(max_items=max_items)
    except ValueError as e:
        log.error(str(e))
        return

    if not user_groups:
        log.info("No user groups found.")
        return

    table = tabulate.tabulate(
        [(ug.id, ug.name) for ug in user_groups], headers=["ID", "Name"],
    )
    rprint(table)


@user_group_cli.command(
    name="get",
    cls=AnyscaleCommand,
    example=command_examples.USER_GROUP_GET_EXAMPLE,
    is_beta=True,
)
@click.option(
    "--id", required=True, type=str, help="The ID of the user group to retrieve.",
)
def get_user_group(id: str) -> None:  # noqa: A002
    """
    Get a specific user group by ID.
    """
    try:
        user_group = anyscale.user_group.get(id=id)
    except ValueError as e:
        log.error(f"Failed to get user group: {e}")
        return

    details = [
        ("ID", user_group.id),
        ("Name", user_group.name),
        ("Organization ID", user_group.org_id),
        ("Created At", user_group.created_at.strftime("%Y-%m-%d %H:%M:%S UTC")),
        ("Updated At", user_group.updated_at.strftime("%Y-%m-%d %H:%M:%S UTC")),
    ]
    table = tabulate.tabulate(details, tablefmt="plain")
    rprint(table)


@user_group_cli.group("membership", help="Manage user group memberships.")
def membership_cli() -> None:
    pass


@membership_cli.command(
    name="list",
    cls=AnyscaleCommand,
    example=command_examples.USER_GROUP_MEMBERSHIP_LIST_EXAMPLE,
    is_beta=True,
)
@click.option(
    "--output",
    "-o",
    type=click.Path(),
    default=None,
    help="Write JSON output to a file instead of stdout.",
)
def list_memberships(output: Optional[str]) -> None:
    """
    List all user groups with their members.

    Shows each user group and which users are members of that group.

    Output is JSON. Use --output to save to a file.
    """
    client = AnyscaleClient()
    try:
        log.info("Listing user group memberships...")
        response = client.list_user_group_memberships()
        result = response.get("result", response)

        groups = result.get("groups", [])
        simple_output = {}
        for group in groups:
            group_name = group.get("group_name", group.get("group_id", "unknown"))
            members = group.get("members", [])
            simple_output[group_name] = sorted(
                m.get("user_email", "") for m in members if m.get("user_email")
            )

        json_output = json.dumps(simple_output, indent=2)

        if output:
            with open(output, "w") as f:
                f.write(json_output)
            log.info(f"Results written to {output}")
        else:
            print(json_output)
    except click.ClickException:
        raise
    except Exception as e:  # noqa: BLE001
        log.error(f"Failed to list user group memberships: {e}")
        raise click.ClickException(str(e))



from typing import Dict, List, Optional, Union

from anyscale._private.models.model_base import ResultIterator
from anyscale._private.sdk import sdk_command
from anyscale.service._private.service_sdk import PrivateServiceSDK
from anyscale.service.models import (
    ServiceConfig,
    ServiceLogMode,
    ServiceSortField,
    ServiceSortOrder,
    ServiceState,
    ServiceStatus,
)


_SERVICE_SDK_SINGLETON_KEY = "service_sdk"

_DEPLOY_EXAMPLE = """
import anyscale
from anyscale.service.models import ServiceConfig

anyscale.service.deploy(
    ServiceConfig(
        name="my-service",
        applications=[
            {"import_path": "main:app"},
        ],
        working_dir=".",
    ),
    canary_percent=50,
)
"""

_DEPLOY_ARG_DOCSTRINGS = {
    "configs": "The config options defining the service.",
    "in_place": "Perform an in-place upgrade without starting a new cluster. This can be used for faster iteration during development but is *not* currently recommended for production deploys. This *cannot* be used to change cluster-level options such as image and compute config (they will be ignored).",
    "canary_percent": "The percentage of traffic to send to the canary version of the service (0-100). This can be used to manually shift traffic toward (or away from) the canary version. If not provided, traffic will be shifted incrementally toward the canary version until it reaches 100. Not supported when using --in-place. This is ignored when restarting a service or creating a new service.",
    "max_surge_percent": "Amount of excess capacity allowed to be used while updating the service (0-100). Defaults to 100. Not supported when using --in-place.",
    "versions": "Enable multi-version deployment by providing a JSON array of objects or a JSON object in text format. Defines the version name, traffic and capacity percents per version. Capacity defaults to traffic.",
    "name": "Name of the service. When running in a workspace, this defaults to the workspace name.",
    "cloud": "The Anyscale Cloud of this workload. If not provided, the organization default will be used (or, if running in a workspace, the cloud of the workspace).",
    "project": "Named project to use for the service. If not provided, the default project for the cloud will be used (or, if running in a workspace, the project of the workspace).",
}


@sdk_command(
    _SERVICE_SDK_SINGLETON_KEY,
    PrivateServiceSDK,
    doc_py_example=_DEPLOY_EXAMPLE,
    arg_docstrings=_DEPLOY_ARG_DOCSTRINGS,
)
def deploy(
    configs: Union[ServiceConfig, List[ServiceConfig]],
    *,
    in_place: bool = False,
    canary_percent: Optional[int] = None,
    max_surge_percent: Optional[int] = None,
    versions: Optional[str] = None,
    name: Optional[str] = None,
    cloud: Optional[str] = None,
    project: Optional[str] = None,
    _private_sdk: Optional[PrivateServiceSDK] = None,
) -> str:
    """Deploy a service.

    If no service with the provided name is running, one will be created, else the existing service will be updated.

    This command is asynchronous, so it always returns immediately.

    Returns the id of the deployed service.
    """
    return _private_sdk.deploy(  # type: ignore
        configs=configs,
        in_place=in_place,
        canary_percent=canary_percent,
        max_surge_percent=max_surge_percent,
        versions=versions,
        name=name,
        cloud=cloud,
        project=project,
    )


_ROLLBACK_EXAMPLE = """
import anyscale

anyscale.service.rollback(name="my-service")
"""

_ROLLBACK_ARG_DOCSTRINGS = {
    "name": "Name of the service. When running in a workspace, this defaults to the workspace name.",
    "cloud": "The Anyscale Cloud of this workload. If not provided, the organization default will be used (or, if running in a workspace, the cloud of the workspace).",
    "project": "Named project to use for the service. If not provided, the default project for the cloud will be used (or, if running in a workspace, the project of the workspace).",
    "max_surge_percent": "Amount of excess capacity allowed to be used while rolling back to the primary version of the service (0-100). Defaults to 100.",
}


@sdk_command(
    _SERVICE_SDK_SINGLETON_KEY,
    PrivateServiceSDK,
    doc_py_example=_ROLLBACK_EXAMPLE,
    arg_docstrings=_ROLLBACK_ARG_DOCSTRINGS,
)
def rollback(
    name: Optional[str],
    *,
    cloud: Optional[str] = None,
    project: Optional[str] = None,
    max_surge_percent: Optional[int] = None,
    _private_sdk: Optional[PrivateServiceSDK] = None,
) -> str:
    """Rollback to the primary version of the service.

    This command can only be used when there is an active rollout in progress. The
    rollout will be cancelled and the service will revert to the primary version.

    This command is asynchronous, so it always returns immediately.

    Returns the id of the rolled back service.
    """
    return _private_sdk.rollback(  # type: ignore
        name=name, cloud=cloud, project=project, max_surge_percent=max_surge_percent
    )


_TERMINATE_EXAMPLE = """
import anyscale

anyscale.service.terminate(name="my-service")
"""

_TERMINATE_ARG_DOCSTRINGS = {
    "id": "ID of the service.",
    "name": "Name of the service. When running in a workspace, this defaults to the workspace name.",
    "cloud": "The Anyscale Cloud of this workload. If not provided, the organization default will be used (or, if running in a workspace, the cloud of the workspace).",
    "project": "Named project to use for the service. If not provided, the default project for the cloud will be used (or, if running in a workspace, the project of the workspace).",
}


@sdk_command(
    _SERVICE_SDK_SINGLETON_KEY,
    PrivateServiceSDK,
    doc_py_example=_TERMINATE_EXAMPLE,
    arg_docstrings=_TERMINATE_ARG_DOCSTRINGS,
)
def terminate(
    id: Optional[str] = None,  # noqa: A002
    name: Optional[str] = None,
    *,
    cloud: Optional[str] = None,
    project: Optional[str] = None,
    _private_sdk: Optional[PrivateServiceSDK] = None,
) -> str:
    """Terminate a service.

    This command is asynchronous, so it always returns immediately.

    Returns the id of the terminated service.
    """
    return _private_sdk.terminate(id=id, name=name, cloud=cloud, project=project)  # type: ignore


_ARCHIVE_EXAMPLE = """
import anyscale

anyscale.service.archive(name="my-service")
"""

_ARCHIVE_ARG_DOCSTRINGS = {
    "id": "ID of the service.",
    "name": "Name of the service. When running in a workspace, this defaults to the workspace name.",
    "cloud": "The Anyscale Cloud of this workload. If not provided, the organization default will be used (or, if running in a workspace, the cloud of the workspace).",
    "project": "Named project to use for the service. If not provided, the default project for the cloud will be used (or, if running in a workspace, the project of the workspace).",
}


@sdk_command(
    _SERVICE_SDK_SINGLETON_KEY,
    PrivateServiceSDK,
    doc_py_example=_ARCHIVE_EXAMPLE,
    arg_docstrings=_ARCHIVE_ARG_DOCSTRINGS,
)
def archive(
    id: Optional[str] = None,  # noqa: A002
    name: Optional[str] = None,
    *,
    cloud: Optional[str] = None,
    project: Optional[str] = None,
    _private_sdk: Optional[PrivateServiceSDK] = None,
) -> str:
    """Archive a service.

    This command is asynchronous, so it always returns immediately.

    Returns the ID of the archived service.
    """
    return _private_sdk.archive(id=id, name=name, cloud=cloud, project=project)  # type: ignore


_DELETE_EXAMPLE = """
import anyscale

anyscale.service.delete(name="my-service")
"""

_DELETE_ARG_DOCSTRINGS = {
    "id": "ID of the service.",
    "name": "Name of the service. When running in a workspace, this defaults to the workspace name.",
    "cloud": "The Anyscale Cloud of this workload. If not provided, the organization default will be used (or, if running in a workspace, the cloud of the workspace).",
    "project": "Named project to use for the service. If not provided, the default project for the cloud will be used (or, if running in a workspace, the project of the workspace).",
}


@sdk_command(
    _SERVICE_SDK_SINGLETON_KEY,
    PrivateServiceSDK,
    doc_py_example=_DELETE_EXAMPLE,
    arg_docstrings=_DELETE_ARG_DOCSTRINGS,
)
def delete(
    id: Optional[str] = None,  # noqa: A002
    name: Optional[str] = None,
    *,
    cloud: Optional[str] = None,
    project: Optional[str] = None,
    _private_sdk: Optional[PrivateServiceSDK] = None,
) -> str:
    """Delete a service.

    This command is asynchronous, so it always returns immediately.

    Returns the ID of the deleted service.
    """
    return _private_sdk.delete(id=id, name=name, cloud=cloud, project=project)  # type: ignore


_STATUS_EXAMPLE = """
import anyscale
from anyscale.service.models import ServiceStatus

status: ServiceStatus = anyscale.service.status(name="my-service")
"""

_STATUS_ARG_DOCSTRINGS = {
    "name": "Name of the service.",
    "cloud": "The Anyscale Cloud of this workload. If not provided, the organization default will be used (or, if running in a workspace, the cloud of the workspace).",
    "project": "Named project to use for the service. If not provided, the default project for the cloud will be used (or, if running in a workspace, the project of the workspace).",
}


@sdk_command(
    _SERVICE_SDK_SINGLETON_KEY,
    PrivateServiceSDK,
    doc_py_example=_STATUS_EXAMPLE,
    arg_docstrings=_STATUS_ARG_DOCSTRINGS,
)
def status(
    name: str,
    *,
    cloud: Optional[str] = None,
    project: Optional[str] = None,
    _private_sdk: Optional[PrivateServiceSDK] = None,
) -> ServiceStatus:
    """Get the status of a service."""
    return _private_sdk.status(name=name, cloud=cloud, project=project)  # type: ignore


_WAIT_EXAMPLE = """
import anyscale
from anyscale.service.models import ServiceState

anyscale.service.wait(name="my-service", state=ServiceState.RUNNING)
"""

_WAIT_ARG_DOCSTRINGS = {
    "name": "Name of the service.",
    "cloud": "The Anyscale Cloud of this workload. If not provided, the organization default will be used (or, if running in a workspace, the cloud of the workspace).",
    "project": "Named project to use for the service. If not provided, the default project for the cloud will be used (or, if running in a workspace, the project of the workspace).",
    "state": "The state to wait for the service to reach.",
    "timeout_s": "Timeout to wait for the service to reach the target state.",
}


@sdk_command(
    _SERVICE_SDK_SINGLETON_KEY,
    PrivateServiceSDK,
    doc_py_example=_WAIT_EXAMPLE,
    arg_docstrings=_WAIT_ARG_DOCSTRINGS,
)
def wait(
    name: str,
    *,
    cloud: Optional[str] = None,
    project: Optional[str] = None,
    state: Union[str, ServiceState] = ServiceState.RUNNING,
    timeout_s: float = 600,
    _private_sdk: Optional[PrivateServiceSDK] = None,
    _interval_s: float = 5,
):
    """Wait for a service to reach a target state."""
    _private_sdk.wait(  # type: ignore
        name=name,
        cloud=cloud,
        project=project,
        state=ServiceState(state),
        timeout_s=timeout_s,
        interval_s=_interval_s,
    )  # type: ignore


_CONTROLLER_LOGS_EXAMPLE = """
import anyscale

anyscale.service.controller_logs("my-service", canary=True)
"""

_CONTROLLER_LOGS_ARG_DOCSTRINGS = {
    "name": "Name of the service.",
    "cloud": "The Anyscale Cloud of this workload. If not provided, the organization default will be used (or, if running in a workspace, the cloud of the workspace).",
    "project": "Named project to use for the service. If not provided, the default project for the cloud will be used (or, if running in a workspace, the project of the workspace).",
    "canary": "Whether to show the logs of the canary version of the service. If not provided, the primary version logs will be shown.",
    "mode": "The mode of log fetching to be used. Supported modes can be found in ServiceLogMode. If not provided, ServiceLogMode.TAIL will be used.",
    "max_lines": "The number of log lines to be fetched. If not provided, 1000 lines will be fetched.",
}

# SDK docs for Service tag operations
_TAGS_ADD_EXAMPLE = """
import anyscale

anyscale.service.add_tags(id="svc_123", tags={"team": "mlops", "env": "prod"})
"""

_TAGS_ADD_ARG_DOCSTRINGS = {
    "id": "ID of the service. Provide either id or name.",
    "name": "Name of the service. Provide either id or name.",
    "cloud": "Cloud name (used when resolving by name).",
    "project": "Project name (used when resolving by name).",
    "tags": "Key/value tags to upsert as a map {key: value}.",
}

_TAGS_REMOVE_EXAMPLE = """
import anyscale

anyscale.service.remove_tags(id="svc_123", keys=["team", "env"])
"""

_TAGS_REMOVE_ARG_DOCSTRINGS = {
    "id": "ID of the service. Provide either id or name.",
    "name": "Name of the service. Provide either id or name.",
    "cloud": "Cloud name (used when resolving by name).",
    "project": "Project name (used when resolving by name).",
    "keys": "List of tag keys to remove.",
}


@sdk_command(
    _SERVICE_SDK_SINGLETON_KEY,
    PrivateServiceSDK,
    doc_py_example=_TAGS_ADD_EXAMPLE,
    arg_docstrings=_TAGS_ADD_ARG_DOCSTRINGS,
)
def add_tags(
    *,
    id: Optional[str] = None,  # noqa: A002
    name: Optional[str] = None,
    cloud: Optional[str] = None,
    project: Optional[str] = None,
    tags: Dict[str, str],
    _private_sdk: Optional[PrivateServiceSDK] = None,
):
    """Upsert (add/update) tag key/value pairs for a service."""
    return _private_sdk.add_tags(  # type: ignore
        id=id, name=name, cloud=cloud, project=project, tags=tags
    )


@sdk_command(
    _SERVICE_SDK_SINGLETON_KEY,
    PrivateServiceSDK,
    doc_py_example=_TAGS_REMOVE_EXAMPLE,
    arg_docstrings=_TAGS_REMOVE_ARG_DOCSTRINGS,
)
def remove_tags(
    *,
    id: Optional[str] = None,  # noqa: A002
    name: Optional[str] = None,
    cloud: Optional[str] = None,
    project: Optional[str] = None,
    keys: List[str],
    _private_sdk: Optional[PrivateServiceSDK] = None,
):
    """Remove tags by key from a service."""
    return _private_sdk.remove_tags(  # type: ignore
        id=id, name=name, cloud=cloud, project=project, keys=keys
    )


_TAGS_LIST_EXAMPLE = """
import anyscale

tags: dict[str, str] = anyscale.service.list_tags(name="my-service")
"""

_TAGS_LIST_ARG_DOCSTRINGS = {
    "id": "ID of the service. Provide either id or name.",
    "name": "Name of the service. Provide either id or name.",
    "cloud": "Cloud name (used when resolving by name).",
    "project": "Project name (used when resolving by name).",
}


@sdk_command(
    _SERVICE_SDK_SINGLETON_KEY,
    PrivateServiceSDK,
    doc_py_example=_TAGS_LIST_EXAMPLE,
    arg_docstrings=_TAGS_LIST_ARG_DOCSTRINGS,
)
def list_tags(
    *,
    id: Optional[str] = None,  # noqa: A002
    name: Optional[str] = None,
    cloud: Optional[str] = None,
    project: Optional[str] = None,
    _private_sdk: Optional[PrivateServiceSDK] = None,
) -> Dict[str, str]:
    """List tags for a service as a key/value mapping."""
    return _private_sdk.list_tags(  # type: ignore
        id=id, name=name, cloud=cloud, project=project
    )


# This is a private command that is not exposed to the user.
@sdk_command(
    _SERVICE_SDK_SINGLETON_KEY,
    PrivateServiceSDK,
    doc_py_example=_CONTROLLER_LOGS_EXAMPLE,
    arg_docstrings=_CONTROLLER_LOGS_ARG_DOCSTRINGS,
)
def _controller_logs(
    name: str,
    *,
    cloud: Optional[str] = None,
    project: Optional[str] = None,
    canary: bool = False,
    mode: Union[str, ServiceLogMode] = ServiceLogMode.TAIL,
    max_lines: int = 1000,
    _private_sdk: Optional[PrivateServiceSDK] = None,
):
    """Wait for a service to reach a target state."""
    return _private_sdk.controller_logs(  # type: ignore
        name,
        cloud=cloud,
        project=project,
        canary=canary,
        mode=mode,
        max_lines=max_lines,
    )


_LIST_EXAMPLE = """
import anyscale
from anyscale.service.models import ServiceState

# Example: Get the first 50 running services
for svc in anyscale.service.list(max_items=50, state_filter=[ServiceState.RUNNING]):
    print(svc.name)
"""

_LIST_ARG_DOCSTRINGS = {
    "service_id": (
        "If provided, returns just the service with this ID "
        "wrapped in a one-page iterator."
    ),
    "name": "Substring to match against the service name.",
    "state_filter": (
        "List of states to include. "
        "May be `ServiceState` enums or case-insensitive strings."
    ),
    "creator_id": "Filter services by user ID.",
    "cloud": "Name of the Anyscale Cloud to search in.",
    "project": "Name of the Anyscale Project to search in.",
    "include_archived": "Include archived services (default: False).",
    "tags_filter": r"Filter services by tags as a dict: \{key: [values...]\}. Values with the same key are ORed; keys are ANDed.",
    # Paging
    "max_items": "Maximum **total** number of items to yield (default: iterate all).",
    "page_size": "Number of items to fetch per API request (default: API default).",
    # Sorting
    "sort_field": "Field to sort by (`NAME`, `STATUS`, `CREATED_AT`).",
    "sort_order": "Sort direction (`ASC` or `DESC`).",
}


# Public command
@sdk_command(
    _SERVICE_SDK_SINGLETON_KEY,
    PrivateServiceSDK,
    doc_py_example=_LIST_EXAMPLE,
    arg_docstrings=_LIST_ARG_DOCSTRINGS,
)
def list(  # noqa: A001, PLR0913
    *,
    # Single-item lookup
    service_id: Optional[str] = None,
    # Filters
    name: Optional[str] = None,
    state_filter: Optional[Union[List[ServiceState], List[str]]] = None,
    creator_id: Optional[str] = None,
    cloud: Optional[str] = None,
    project: Optional[str] = None,
    include_archived: bool = False,
    tags_filter: Optional[Dict[str, List[str]]] = None,
    # Paging
    max_items: Optional[int] = None,
    page_size: Optional[int] = None,
    # Sorting
    sort_field: Optional[Union[str, ServiceSortField]] = None,
    sort_order: Optional[Union[str, ServiceSortOrder]] = None,
    # Injected SDK
    _private_sdk: Optional[PrivateServiceSDK] = None,
) -> ResultIterator[ServiceStatus]:
    """List services or fetch a single service by ID."""
    return _private_sdk.list(  # type: ignore
        service_id=service_id,
        name=name,
        state_filter=state_filter,
        creator_id=creator_id,
        cloud=cloud,
        project=project,
        include_archived=include_archived,
        tags_filter=tags_filter,
        max_items=max_items,
        page_size=page_size,
        sort_field=sort_field,
        sort_order=sort_order,
    )



from typing import List, Optional

from anyscale._private.models.model_base import ResultIterator
from anyscale._private.sdk import sdk_command
from anyscale.cloud._private.cloud_sdk import Cloud, PrivateCloudSDK
from anyscale.cloud.models import CreateCloudCollaborator


_CLOUD_SDK_SINGLETON_KEY = "cloud_sdk"

_LIST_EXAMPLE = """
import anyscale

# Example: Get the first 50 clouds
for cloud in anyscale.cloud.list(max_items=50):
    print(cloud.name)
"""

_LIST_ARG_DOCSTRINGS = {
    "cloud_id": "If provided, returns just the cloud with this ID wrapped in a one-page iterator.",
    "name": "Substring or exact name to match against the cloud name.",
    "max_items": "Maximum total number of items to yield (default: iterate all).",
    "page_size": "Number of items to fetch per API request (default: API default).",
}


@sdk_command(
    _CLOUD_SDK_SINGLETON_KEY,
    PrivateCloudSDK,
    doc_py_example=_LIST_EXAMPLE,
    arg_docstrings=_LIST_ARG_DOCSTRINGS,
)
def list(  # noqa: A001
    *,
    cloud_id: Optional[str] = None,
    name: Optional[str] = None,
    max_items: Optional[int] = None,
    page_size: Optional[int] = None,
    _private_sdk: Optional[PrivateCloudSDK] = None,
) -> ResultIterator[Cloud]:
    """List clouds or fetch a single cloud by ID/name."""
    return _private_sdk.list(  # type: ignore
        cloud_id=cloud_id, name=name, max_items=max_items, page_size=page_size
    )


_ADD_COLLABORATORS_EXAMPLE = """
import anyscale
from anyscale.cloud.models import CloudPermissionLevel, CreateCloudCollaborator

anyscale.cloud.add_collaborators(
    cloud="cloud_name",
    collaborators=[
        CreateCloudCollaborator(
            email="test1@anyscale.com",
            permission_level=CloudPermissionLevel.WRITE,
        ),
        CreateCloudCollaborator(
            email="test2@anyscale.com",
            permission_level=CloudPermissionLevel.READONLY,
        ),
    ],
)
"""

_ADD_COLLABORATORS_ARG_DOCSTRINGS = {
    "cloud": "The cloud to add users to.",
    "collaborators": "The list of collaborators to add to the cloud.",
}


@sdk_command(
    _CLOUD_SDK_SINGLETON_KEY,
    PrivateCloudSDK,
    doc_py_example=_ADD_COLLABORATORS_EXAMPLE,
    arg_docstrings=_ADD_COLLABORATORS_ARG_DOCSTRINGS,
)
def add_collaborators(
    cloud: str,
    collaborators: List[CreateCloudCollaborator],
    *,
    _private_sdk: Optional[PrivateCloudSDK] = None,
) -> str:
    """
    Batch add collaborators to a cloud.

    :param cloud: The cloud to add users to.
    :param collaborators: The list of collaborators to add to the cloud.
    """
    return _private_sdk.add_collaborators(cloud, collaborators)  # type: ignore


_GET_EXAMPLE = """
import anyscale

# Get a cloud by ID
cloud_by_id = anyscale.cloud.get(id="cloud_id")

# Get a cloud by name
cloud_by_name = anyscale.cloud.get(name="cloud_name")
"""

_GET_ARG_DOCSTRINGS = {
    "id": "The ID of the cloud to retrieve.",
    "name": "The name of the cloud to retrieve.",
}


@sdk_command(
    _CLOUD_SDK_SINGLETON_KEY,
    PrivateCloudSDK,
    doc_py_example=_GET_EXAMPLE,
    arg_docstrings=_GET_ARG_DOCSTRINGS,
)
def get(
    id: Optional[str] = None,  # noqa: A002
    name: Optional[str] = None,
    *,
    _private_sdk: Optional[PrivateCloudSDK] = None,
) -> Optional[Cloud]:
    """
    Get the cloud model for the provided cloud ID or name.

    If neither ID nor name is provided, returns `None`.

    :param id: The ID of the cloud to retrieve.
    :param name: The name of the cloud to retrieve.
    :return: A `Cloud` object if found, otherwise `None`.
    """
    return _private_sdk.get(id=id, name=name)  # type: ignore


_GET_DEFAULT_EXAMPLE = """
import anyscale

# Get the user's default cloud
default_cloud = anyscale.cloud.get_default()
"""


@sdk_command(
    _CLOUD_SDK_SINGLETON_KEY,
    PrivateCloudSDK,
    doc_py_example=_GET_DEFAULT_EXAMPLE,
    arg_docstrings={},
)
def get_default(*, _private_sdk: Optional[PrivateCloudSDK] = None) -> Optional[Cloud]:
    """
    Get the user's default cloud.

    :return: The default `Cloud` object if it exists, otherwise `None`.
    """
    return _private_sdk.get_default()  # type: ignore


_TERMINATE_SYSTEM_CLUSTER_EXAMPLE = """
import anyscale

# Terminate the system cluster for the cloud with the specified ID
anyscale.cloud.terminate_system_cluster(cloud_id="cloud_id", wait=True)
"""

_TERMINATE_SYSTEM_CLUSTER_ARG_DOCSTRINGS = {
    "cloud_id": "The ID of the cloud whose system cluster should be terminated.",
    "wait": "If True, wait for the system cluster to be terminated before returning. Defaults to False.",
}


@sdk_command(
    _CLOUD_SDK_SINGLETON_KEY,
    PrivateCloudSDK,
    doc_py_example=_TERMINATE_SYSTEM_CLUSTER_EXAMPLE,
    arg_docstrings=_TERMINATE_SYSTEM_CLUSTER_ARG_DOCSTRINGS,
)
def terminate_system_cluster(
    cloud_id: str,
    wait: Optional[bool] = False,
    *,
    _private_sdk: Optional[PrivateCloudSDK] = None,
) -> str:
    """
    Terminate the system cluster for the specified cloud.

    :param cloud: The name of the cloud whose system cluster should be terminated.
    :param wait: If True, wait for the system cluster to be terminated before returning. Defaults to False.
    :return: ID of the terminated system cluster.
    """
    return _private_sdk.terminate_system_cluster(cloud_id, wait)  # type: ignore



from typing import List, Optional

from anyscale._private.anyscale_client import AnyscaleClientInterface
from anyscale._private.sdk import sdk_docs
from anyscale._private.sdk.base_sdk import Timer
from anyscale.cli_logger import BlockLogger
from anyscale.resource_quota._private.resource_quota_sdk import PrivateResourceQuotaSDK
from anyscale.resource_quota.commands import (
    _CREATE_DOCSTRINGS,
    _CREATE_EXAMPLE,
    _DELETE_DOCSTRINGS,
    _DELETE_EXAMPLE,
    _DISABLE_DOCSTRINGS,
    _DISABLE_EXAMPLE,
    _ENABLE_DOCSTRINGS,
    _ENABLE_EXAMPLE,
    _LIST_DOCSTRINGS,
    _LIST_EXAMPLE,
    create as create,
    delete as delete,
    disable as disable,
    enable as enable,
    list as list,  # noqa: A004 - claude_comment("claude-opus-4-5", "SDK public API re-export")
)
from anyscale.resource_quota.models import CreateResourceQuota, ResourceQuota


class ResourceQuotaSDK:
    def __init__(
        self,
        *,
        client: Optional[AnyscaleClientInterface] = None,
        logger: Optional[BlockLogger] = None,
        timer: Optional[Timer] = None,
    ):
        self._private_sdk = PrivateResourceQuotaSDK(
            client=client, logger=logger, timer=timer
        )

    @sdk_docs(
        doc_py_example=_CREATE_EXAMPLE, arg_docstrings=_CREATE_DOCSTRINGS,
    )
    def create(  # noqa: F811
        self, create_resource_quota: CreateResourceQuota,
    ) -> ResourceQuota:
        """Create a resource quota.
        """
        return self._private_sdk.create(create_resource_quota)

    @sdk_docs(
        doc_py_example=_LIST_EXAMPLE, arg_docstrings=_LIST_DOCSTRINGS,
    )
    def list(  # noqa: F811
        self,
        name: Optional[str] = None,
        cloud: Optional[str] = None,
        creator_id: Optional[str] = None,
        is_enabled: Optional[bool] = None,
        max_items: int = 20,
    ) -> List[ResourceQuota]:
        """List resource quotas.
        """
        return self._private_sdk.list(
            name=name,
            cloud=cloud,
            creator_id=creator_id,
            is_enabled=is_enabled,
            max_items=max_items,
        )

    @sdk_docs(
        doc_py_example=_DELETE_EXAMPLE, arg_docstrings=_DELETE_DOCSTRINGS,
    )
    def delete(  # noqa: F811
        self, resource_quota_id: str,
    ):
        """Delete a resource quota.
        """
        return self._private_sdk.delete(resource_quota_id)

    @sdk_docs(
        doc_py_example=_ENABLE_EXAMPLE, arg_docstrings=_ENABLE_DOCSTRINGS,
    )
    def enable(  # noqa: F811
        self, resource_quota_id: str,
    ):
        """Enable a resource quota.
        """
        return self._private_sdk.set_status(resource_quota_id, True)

    @sdk_docs(
        doc_py_example=_DISABLE_EXAMPLE, arg_docstrings=_DISABLE_DOCSTRINGS,
    )
    def disable(  # noqa: F811
        self, resource_quota_id: str,
    ):
        """Disable a resource quota.
        """
        return self._private_sdk.set_status(resource_quota_id, False)



import asyncio
import copy
import json
from typing import Any, Dict, List, Optional, Tuple, Union

from anyscale._private.models.model_base import ResultIterator
from anyscale._private.workload import WorkloadSDK
from anyscale.client.openapi_client.models.apply_production_service_multi_version_v2_model import (
    ApplyProductionServiceMultiVersionV2Model,
)
from anyscale.client.openapi_client.models.decorated_production_service_v2_api_model import (
    DecoratedProductionServiceV2APIModel,
)
from anyscale.client.openapi_client.models.decorated_production_service_v2_version_api_model import (
    DecoratedProductionServiceV2VersionAPIModel,
)
from anyscale.client.openapi_client.models.decoratedlistserviceapimodel_list_response import (
    DecoratedlistserviceapimodelListResponse,
)
from anyscale.client.openapi_client.models.list_response_metadata import (
    ListResponseMetadata,
)
from anyscale.client.openapi_client.models.resource_tag_resource_type import (
    ResourceTagResourceType,
)
from anyscale.compute_config.models import ComputeConfig
from anyscale.sdk.anyscale_client.models import (
    AccessConfig,
    ApplyProductionServiceV2Model,
    GrpcProtocolConfig,
    ProductionServiceV2VersionModel,
    Protocols,
    RayGCSExternalStorageConfig as APIRayGCSExternalStorageConfig,
    RolloutStrategy,
    ServiceConfig as ExternalAPIServiceConfig,
    ServiceEventCurrentState,
    ServiceSortField,
    TracingConfig as APITracingConfg,
)
from anyscale.service.models import (
    RayGCSExternalStorageConfig,
    ServiceConfig,
    ServiceLogMode,
    ServiceSortOrder,
    ServiceState,
    ServiceStatus,
    ServiceVersionStatus,
    TracingConfig,
)
from anyscale.shared_anyscale_utils.constants import SERVICE_VERSION_ID_TRUNCATED_LEN
from anyscale.utils.runtime_env import parse_requirements_file
from anyscale.utils.workspace_notification import (
    WorkspaceNotification,
    WorkspaceNotificationAction,
)


MAX_PAGE_SIZE = 50


class PrivateServiceSDK(WorkloadSDK):
    def _override_application_runtime_envs(
        self,
        config: ServiceConfig,
        *,
        autopopulate_in_workspace: bool = True,
        cloud_id: Optional[str] = None,
        workspace_requirements_path: Optional[str] = None,
    ) -> ServiceConfig:
        """Overrides the runtime_env of each application in the config.

        Local directories specified in the 'working_dir' or 'py_modules' fields will be
        uploaded and replaced with the resulting remote URIs.

        Requirements files will be loaded and populated into the 'pip' field.

        If autopopulate_from_workspace is passed and this code is running inside a
        workspace, the following defaults will be applied:
            - 'working_dir' will be set to '.'.
            - 'pip' will be set to the workspace-managed requirements file.
        """
        new_applications = copy.deepcopy(config.applications)
        new_runtime_envs = [app.get("runtime_env", {}) for app in new_applications]

        # Multi-deployment compute configs are not supported for services.
        new_runtime_envs = self.override_and_upload_local_dirs_single_deployment(
            new_runtime_envs,
            working_dir_override=config.working_dir,
            excludes_override=config.excludes,
            cloud_id=cloud_id,
            autopopulate_in_workspace=autopopulate_in_workspace,
            additional_py_modules=config.py_modules,
            py_executable_override=config.py_executable,
        )
        new_runtime_envs = self.override_and_load_requirements_files(
            new_runtime_envs,
            requirements_override=config.requirements,
            workspace_requirements_path=workspace_requirements_path,
        )
        new_runtime_envs = self.update_env_vars(
            new_runtime_envs, env_vars_updates=config.env_vars
        )

        for app, new_runtime_env in zip(new_applications, new_runtime_envs):
            if new_runtime_env:
                app["runtime_env"] = new_runtime_env

        return config.options(
            applications=new_applications,
            requirements=None,
            working_dir=None,
            excludes=None,
        )

    def _get_default_name(self) -> str:
        """Get a default name for the service.

        A default is currently only generated when running inside a workspace
        (from the workspace cluster name), so this function errors if called outside
        a workspace.
        """
        name = self.get_current_workspace_name()
        if name is None:
            raise ValueError(
                "A service name must be provided when running outside of a workspace."
            )

        self.logger.info(f"No name was specified, using default: '{name}'.")
        return name

    def _log_deployed_service_info(
        self,
        service: DecoratedProductionServiceV2APIModel,
        *,
        canary_percent: Optional[int],
    ):
        """Log user-facing information about a deployed service."""
        if not service.primary_version:
            version_info = "versions: {versions}".format(
                versions=[v.version for v in service.versions]
            )
        else:
            version_info = "version ID: {version_id}".format(
                version_id=self._get_user_facing_service_version_id(
                    service.canary_version
                    if service.canary_version is not None
                    else service.primary_version
                )
            )
        details = (
            "("
            + version_info
            + (
                ")"
                if canary_percent is None
                else f", target canary percent: {canary_percent})"
            )
        )

        message = f"Service '{service.name}' deployed {details}."
        self.logger.info(message)
        if self.client.inside_workspace():
            self.client.send_workspace_notification(
                WorkspaceNotification(
                    body=message,
                    action=WorkspaceNotificationAction(
                        type="navigate-service", title="View Service", value=service.id,
                    ),
                ),
            )

        self.logger.info(
            f"View the service in the UI: '{self.client.get_service_ui_url(service.id)}'"
        )
        self.logger.info(
            "Query the service once it's running using the following curl command (add the path you want to query):"
        )
        auth_token_header = (
            ""
            if service.auth_token is None
            else f'-H "Authorization: Bearer {service.auth_token}" '
        )
        self.logger.info(f"curl {auth_token_header}{service.base_url}/")

    def _build_ray_serve_config(self, config: ServiceConfig) -> Dict[str, Any]:
        ray_serve_config: Dict[str, Any] = {"applications": config.applications}
        if config.http_options:
            ray_serve_config["http_options"] = config.http_options
        if config.grpc_options:
            ray_serve_config["grpc_options"] = config.grpc_options
        if config.logging_config:
            ray_serve_config["logging_config"] = config.logging_config

        return ray_serve_config

    def _build_grpc_protocol_config(self, config: ServiceConfig) -> GrpcProtocolConfig:
        enabled = (
            config.grpc_options is not None
            and len(config.grpc_options.get("service_names", [])) > 0
        )
        service_names = (
            config.grpc_options.get("service_names", [])
            if config.grpc_options is not None
            else []
        )
        # Since config doesn't allow passing a port, keeping it as default port.
        return GrpcProtocolConfig(enabled=enabled, service_names=service_names)

    def _build_apply_service_model_for_in_place_update(  # noqa: PLR0912
        self,
        name: str,
        config: ServiceConfig,
        *,
        canary_percent: Optional[int] = None,
        max_surge_percent: Optional[int] = None,
        existing_service: Optional[DecoratedProductionServiceV2APIModel] = None,
    ) -> ApplyProductionServiceV2Model:
        """Build the ApplyProductionServiceV2Model for an in_place update.

        in_place updates:
            - must be performed on an existing service.
            - cannot be performed while a rollout is active.
            - do not support rollout options (canary_percent and max_surge_percent).
            - do not support changing cluster-level config options (image, compute_config, etc.).

        If cluster-level options are provided, a warning will be logged and they will be ignored.
        """
        if canary_percent is not None:
            raise ValueError(
                "canary_percent cannot be specified when doing an in_place update."
            )
        if max_surge_percent is not None:
            raise ValueError(
                "max_surge_percent cannot be specified when doing an in_place update."
            )
        if existing_service is None:
            raise RuntimeError(
                "in_place can only be used to update running services, "
                f"but no service was found with name '{name}'."
            )
        if existing_service.current_state in {
            ServiceEventCurrentState.TERMINATING,
            ServiceEventCurrentState.TERMINATED,
        }:
            raise RuntimeError(
                "in_place can only be used to update running services, "
                f"but service '{name}' is terminated."
            )

        if existing_service.canary_version is not None:
            raise RuntimeError(
                "in_place updates cannot be used while a rollout is in progress. "
                "Complete the rollout or roll back first."
            )

        if any(
            [
                config.image_uri is not None,
                config.registry_login_secret is not None,
                config.containerfile is not None,
                config.compute_config is not None,
                config.ray_gcs_external_storage_config is not None,
                config.tracing_config is not None,
            ]
        ):
            self.logger.warning(
                "Cluster-level options such as image and compute config "
                "are ignored when performing an in_place update."
            )

        existing_config: ProductionServiceV2VersionModel = (
            existing_service.primary_version
        )
        query_auth_token_enabled = existing_service.auth_token is not None
        cloud_id = self.client.get_cloud_id(
            compute_config_id=existing_config.compute_config_id
        )
        config = self._override_application_runtime_envs(
            config,
            cloud_id=cloud_id,
            workspace_requirements_path=self.client.get_workspace_requirements_path(),
        )

        project_id = self.client.get_project_id(
            parent_cloud_id=cloud_id, name=config.project
        )
        return ApplyProductionServiceV2Model(
            name=name,
            version=config.version_name,
            project_id=project_id,
            ray_serve_config=self._build_ray_serve_config(config),
            build_id=existing_config.build_id,
            compute_config_id=existing_config.compute_config_id,
            rollout_strategy="IN_PLACE",
            config=ExternalAPIServiceConfig(
                access=AccessConfig(use_bearer_token=query_auth_token_enabled),
                protocols=Protocols(grpc=self._build_grpc_protocol_config(config)),
                env_vars=config.env_vars,
            ),
            ray_gcs_external_storage_config=existing_config.ray_gcs_external_storage_config,
            tracing_config=existing_config.tracing_config,
            tags=config.tags,
        )

    def _build_apply_service_model_for_rollout(  # noqa: PLR0912
        self,
        name: str,
        config: ServiceConfig,
        rollout_strategy: RolloutStrategy,
        *,
        canary_percent: Optional[int] = None,
        max_surge_percent: Optional[int] = None,
        existing_service: Optional[DecoratedProductionServiceV2APIModel] = None,
        traffic_percent: Optional[int] = None,
    ) -> ApplyProductionServiceV2Model:
        """Build the ApplyProductionServiceV2Model for a rolling update."""

        build_id = None
        if config.containerfile is not None:
            build_id = self._image_sdk.build_image_from_containerfile(
                name=f"image-for-service-{name}",
                containerfile=self.get_containerfile_contents(config.containerfile),
                anonymous=True,
                ray_version=config.ray_version,
            )
        elif config.image_uri is not None:
            build_id = self._image_sdk.registery_image(
                image_uri=config.image_uri,
                registry_login_secret=config.registry_login_secret,
                ray_version=config.ray_version,
            )

        if self._image_sdk.enable_image_build_for_tracked_requirements:
            requirements_path_to_be_populated_in_runtime_env = None
            requirements_path = self.client.get_workspace_requirements_path()
            if requirements_path is not None:
                requirements = parse_requirements_file(requirements_path)
                if requirements:
                    build_id = self._image_sdk.build_image_from_requirements(
                        name=f"image-for-service-{name}",
                        base_build_id=self.client.get_default_build_id(),
                        requirements=requirements,
                    )
        else:
            requirements_path_to_be_populated_in_runtime_env = (
                self.client.get_workspace_requirements_path()
            )

        if build_id is None:
            build_id = self.client.get_default_build_id()

        compute_config_id, cloud_id = self.resolve_compute_config_and_cloud_id(
            compute_config=config.compute_config, cloud=config.cloud
        )

        project_id = self.client.get_project_id(
            parent_cloud_id=cloud_id, name=config.project
        )
        if (
            existing_service is not None
            and existing_service.primary_version is not None
        ):
            existing_cloud_id = self.client.get_cloud_id(
                compute_config_id=existing_service.primary_version.compute_config_id
            )
            if cloud_id != existing_cloud_id:
                new_cloud = self.client.get_cloud(cloud_id=cloud_id)
                assert new_cloud is not None
                existing_cloud = self.client.get_cloud(cloud_id=existing_cloud_id)
                assert existing_cloud is not None
                raise ValueError(
                    f"The cloud for a service cannot be changed once it's created. "
                    f"Service '{name}' was created on cloud '{existing_cloud.name}', but "
                    f"the provided config is for cloud '{new_cloud.name}'."
                )

        env_vars_from_workspace = self.client.get_workspace_env_vars()
        if env_vars_from_workspace:
            if config.env_vars:
                # the precedence should be cli > workspace
                env_vars_from_workspace.update(config.env_vars)
                config = config.options(env_vars=env_vars_from_workspace)
            else:
                config = config.options(env_vars=env_vars_from_workspace)

        config = self._override_application_runtime_envs(
            config,
            cloud_id=cloud_id,
            workspace_requirements_path=requirements_path_to_be_populated_in_runtime_env,
        )

        ray_gcs_external_storage_config = None
        if config.ray_gcs_external_storage_config is not None:
            assert isinstance(
                config.ray_gcs_external_storage_config, RayGCSExternalStorageConfig
            )
            ray_gcs_external_storage_config = APIRayGCSExternalStorageConfig(
                enable=config.ray_gcs_external_storage_config.enabled,
            )
            if config.ray_gcs_external_storage_config.address is not None:
                ray_gcs_external_storage_config.address = (
                    config.ray_gcs_external_storage_config.address
                )
            if config.ray_gcs_external_storage_config.certificate_path is not None:
                ray_gcs_external_storage_config.redis_certificate_path = (
                    config.ray_gcs_external_storage_config.certificate_path
                )

        tracing_config = None
        if config.tracing_config is not None:
            assert isinstance(config.tracing_config, TracingConfig)
            tracing_config = APITracingConfg(enabled=config.tracing_config.enabled,)
            if config.tracing_config.exporter_import_path is not None:
                tracing_config.exporter_import_path = (
                    config.tracing_config.exporter_import_path
                )
            if config.tracing_config.sampling_ratio is not None:
                tracing_config.sampling_ratio = config.tracing_config.sampling_ratio

        return ApplyProductionServiceV2Model(
            name=name,
            version=config.version_name,
            project_id=project_id,
            ray_serve_config=self._build_ray_serve_config(config),
            build_id=build_id,
            compute_config_id=compute_config_id,
            canary_percent=canary_percent,
            max_surge_percent=max_surge_percent,
            rollout_strategy=rollout_strategy,
            config=ExternalAPIServiceConfig(
                access=AccessConfig(use_bearer_token=config.query_auth_token_enabled),
                protocols=Protocols(grpc=self._build_grpc_protocol_config(config)),
                env_vars=config.env_vars,
            ),
            ray_gcs_external_storage_config=ray_gcs_external_storage_config,
            tracing_config=tracing_config,
            traffic_percent=traffic_percent,
            tags=config.tags,
        )

    def _build_apply_service_model_for_multi_version(
        self,
        name: str,
        cloud: Optional[str],
        project: Optional[str],
        configs: List[ServiceConfig],
        existing_service: Optional[DecoratedProductionServiceV2APIModel],
        versions: List[Dict[str, Any]],
    ) -> ApplyProductionServiceV2Model:
        cloud_id = self.client.get_cloud_id(cloud_name=cloud)
        project_id = self.client.get_project_id(parent_cloud_id=cloud_id, name=project)

        # implement a map between version name and config file
        # This way, we can iterate over the version config and corresponding config file to the _build_apply_service_model_for_rollout function
        # Also, maybe add an assertion to confirm the version name is the same as the config file name.
        # I also need to add a validation logic to confirm if all the version config passed(the version name) all has the corresponding config file.
        version_service_map = self._get_service_version_map(configs, versions, existing_service)  # type: ignore
        existing_versions: Dict[str, DecoratedProductionServiceV2VersionAPIModel] = {}
        if existing_service is not None:
            existing_versions = {
                v.version: v
                for v in self.client.get_service_versions(
                    existing_service.id, read_all_versions=True
                )
            }

        service_versions: List[ApplyProductionServiceV2Model] = []
        for _, (config, version) in version_service_map.items():
            version_name = version["name"]
            traffic_percent = version["traffic_percent"]
            if traffic_percent == 0:
                self.logger.warning(
                    f"Setting traffic for version {version_name} to 0 will terminate it."
                )

            # If config is None, it means the version is already deployed, and the user is likely to be only updating the traffic/capacity percent without serve config update.
            if config is None:
                if version_name not in existing_versions:
                    raise ValueError(
                        f"Version {version_name} does not exist in the existing service."
                    )
                existing_version = existing_versions[version_name]

                service_version = ApplyProductionServiceV2Model(
                    name=name,
                    project_id=project_id,
                    ray_serve_config=existing_version.ray_serve_config,
                    build_id=existing_version.build_id,
                    compute_config_id=existing_version.compute_config_id,
                    rollout_strategy=RolloutStrategy.MULTI_VERSION,
                    # TODO(doyoung): "config" is a service level config, so I'm assuming this is not necessary when this model is
                    # representing a service version. If not, may need to add and store 'config' field in the version model, ProductionServiceV2VersionModel,
                    # from the backend when the service version first gets created so it can be read from here for reconstruction.
                    config=None,
                    ray_gcs_external_storage_config=existing_version.ray_gcs_external_storage_config,
                    tracing_config=existing_version.tracing_config,
                    version=version_name,
                    traffic_percent=traffic_percent,
                )
            else:
                service_version = self._build_apply_service_model_for_rollout(
                    name,
                    config,
                    RolloutStrategy.MULTI_VERSION,
                    existing_service=existing_service,
                    traffic_percent=traffic_percent,
                )

            service_versions.append(service_version)

        assert len(service_versions) > 0, "service_versions should not be empty"
        return ApplyProductionServiceMultiVersionV2Model(
            rollout_strategy=RolloutStrategy.MULTI_VERSION,
            cloud_id=cloud_id,
            project_id=project_id,
            service_versions=service_versions,
        )

    def _validate_versions(
        self, versions: str
    ) -> List[Dict[str, Any]]:  # noqa: C901, PLR0912
        # Parse versions JSON string to list of dictionaries
        try:
            versions = json.loads(versions)
        except json.JSONDecodeError as e:
            raise ValueError(f"Invalid JSON format for --versions: {e}")

        for version in versions:
            # traffic_percent must be specified in all versions by the users.
            # If some version does not specify traffic_percent, raise an error.
            if "name" not in version:
                raise ValueError("Name is required for each version.")
            if "traffic_percent" not in version:
                raise ValueError("Traffic percent is required for each version.")

            # Verify the key names are valid. Only either of name, traffic_percent, or capacity_percent are allowed.
            for key in version:
                if key not in [
                    "name",
                    "traffic_percent",
                ]:
                    raise ValueError(f"Invalid key in versions: {key}")

        if isinstance(versions, dict):
            # Convert object form {"v1": 25, "v2": 75} to full list form
            versions = [  # type: ignore
                {"name": name, "traffic_percent": percent}
                for name, percent in versions.items()
            ]
        elif not isinstance(versions, list):
            raise ValueError(
                "--versions must be a JSON array of objects or a JSON object in text format."
            )

        # Verify sum of each traffic_percent from each version is 100.
        traffic_percent_sum = sum(version["traffic_percent"] for version in versions)
        if traffic_percent_sum != 100:
            raise ValueError("Sum of traffic percent for all versions must be 100.")

        return versions  # type: ignore

    def _read_and_validate_configs(
        self, configs: List[ServiceConfig]
    ) -> Tuple[str, Optional[str], Optional[str]]:
        """The service name, cloud, and project must be identical for all versions."""

        # If all the name field from configs are identical, use the first one as the name.
        # If all the name field from configs are not provided, set to None, use the default name.
        if all(config.name is None for config in configs):
            name = self._get_default_name()
        elif all(config.name == configs[0].name for config in configs):
            name = configs[0].name
        else:
            raise ValueError("All provided names in config files must be identical.")

        # If all the cloud field from configs are identical, use the first one as the cloud.
        # If all the cloud field from configs are all set to None, set cloud to None. Default cloud will be used from backend.
        if all(config.cloud is None for config in configs):
            cloud = None
        elif all(config.cloud == configs[0].cloud for config in configs):
            cloud = configs[0].cloud
        else:
            raise ValueError("All provided clouds in config files must be identical.")

        # If all the project field from configs are identical, use the first one as the project.
        # If all the project field from configs are all set to None, set project to None. Default project will be used from backend.
        if all(config.project is None for config in configs):
            project = None
        elif all(config.project == configs[0].project for config in configs):
            project = configs[0].project
        else:
            raise ValueError("All provided projects in config files must be identical.")

        return name, cloud, project

    def _get_service_version_map(
        self,
        configs: List[ServiceConfig],
        versions: Optional[List[Dict[str, Any]]],
        existing_service: Optional[DecoratedProductionServiceV2APIModel],
    ) -> Dict[str, Tuple[ServiceConfig, Dict[str, Any]]]:
        service_config_to_version_config = {}
        if versions is not None:
            if len(configs) > len(versions):
                raise ValueError(
                    "Number of config files passed must be less than or equal to the number of version configs when multiple versions are being deployed."
                )

            for service_config in configs:
                for version_config in versions:
                    if service_config.version_name == version_config["name"]:
                        service_config_to_version_config[version_config["name"]] = (
                            service_config,
                            version_config,
                        )
                        break

                # version name in the ServiceConfig must have a matching version name in the version config.
                # If this point is reached, then, it means there's a config file passed with version name that doesn't exist in the version config.
                if service_config.version_name not in service_config_to_version_config:
                    raise ValueError(
                        f"version_name {service_config.version_name!r} in service config {service_config} does not match any of the version names provided with --versions."
                    )

            # version name in version config doesn't need to have a matching version name in service config. If there's not a matching, this
            # means the version is already deployed, and the user is likely to be only updating the traffic/capacity percent.
            for version_config in versions:
                version_name = version_config["name"]
                if version_name not in service_config_to_version_config:
                    if existing_service is None:
                        raise ValueError(
                            f"--config-file corresponding to version {version_name!r} must be provided if the Service is not already deployed."
                        )
                    # TODO(doyoung): Perhaps add a validaiton logic to confirm if the version name already exists in the existing service through existing_service.versions.
                    service_config_to_version_config[version_name] = (
                        None,
                        version_config,
                    )

        return service_config_to_version_config

    def _get_rollout_strategy(  # noqa: PLR0912
        self,
        existing_service: Optional[DecoratedProductionServiceV2APIModel],
        version_weights: Optional[List[Dict[str, Any]]],
        in_place: bool,
        configs: List[ServiceConfig],
    ) -> RolloutStrategy:
        """Determine the rollout strategy to use.

        If `in_place` is True, return RolloutStrategy.IN_PLACE.
        If `version_weights` is None or if the service is going from 0 or 1 versions to 1 version, return RolloutStrategy.ROLLOUT.
        Otherwise, return RolloutStrategy.MULTI_VERSION.
        """

        active_versions = len(existing_service.versions) if existing_service else 0

        if in_place:
            if version_weights is not None:
                raise ValueError(
                    "In-place updates are not supported for multi-version services."
                )

            return RolloutStrategy.IN_PLACE
        elif version_weights is None or (
            len(version_weights) == 1 and active_versions <= 1
        ):
            if existing_service is None and len(configs) == 0:
                raise ValueError(
                    "A config file must be provided when deploying a new service."
                )

            return RolloutStrategy.ROLLOUT
        else:
            return RolloutStrategy.MULTI_VERSION

    def deploy(  # noqa: PLR0912, C901
        self,
        configs: Union[ServiceConfig, List[ServiceConfig]],
        *,
        in_place: bool = False,
        canary_percent: Optional[int] = None,
        max_surge_percent: Optional[int] = None,
        versions: Optional[str] = None,
        name: Optional[str] = None,
        cloud: Optional[str] = None,
        project: Optional[str] = None,
    ) -> str:
        """Deploy or update a service.

        Args:
            configs: Service configuration(s) to deploy. Can be a single ServiceConfig or a list of ServiceConfig objects for multi-version deployments.
            versions: JSON string specifying version configurations for multi-version deployments.
            name: Name of the service. If not provided, defaults to workspace cluster name.
            cloud: Cloud provider to deploy to.
            project: Project to deploy the service in.

        Returns:
            The service ID.
        """
        if versions is not None:
            versions = self._validate_versions(versions)  # type: ignore

        if isinstance(configs, ServiceConfig):
            if versions is not None:
                raise ValueError(
                    "To deploy with --versions, a list of ServiceConfig must be provided."
                )

            if not isinstance(in_place, bool):
                raise TypeError("in_place must be a bool.")

            if canary_percent is not None:
                if not isinstance(canary_percent, int):
                    raise TypeError("canary_percent must be an int.")
                if canary_percent < 0 or canary_percent > 100:
                    raise ValueError("canary_percent must be between 0 and 100.")

            if max_surge_percent is not None:
                if not isinstance(max_surge_percent, int):
                    raise TypeError("max_surge_percent must be an int.")
                if max_surge_percent < 0 or max_surge_percent > 100:
                    raise ValueError("max_surge_percent must be between 0 and 100.")

            name = configs.name
            cloud = configs.cloud
            project = configs.project
            configs = [configs]

        elif isinstance(configs, List):
            if versions is None:
                raise ValueError(
                    "--versions must be provided when deploying with multiple config files."
                )

            if len(configs) == 0:
                if name is None:
                    raise ValueError(
                        "When using --versions, a service name must be provided if no config files are given."
                    )
            else:
                name, cloud, project = self._read_and_validate_configs(configs)

        else:
            raise ValueError(
                "Single object of ServiceConfig or list of ServiceConfig must be provided."
            )

        if name is None:
            name = self._get_default_name()

        existing_service: Optional[DecoratedProductionServiceV2APIModel] = (
            self.client.get_service(name=name, cloud=cloud, project=project)
        )
        if existing_service is None:
            self.logger.info(f"Starting new service '{name}'.")
        elif existing_service.current_state == ServiceEventCurrentState.TERMINATED:
            self.logger.info(f"Restarting existing service '{name}'.")
        else:
            self.logger.info(f"Updating existing service '{name}'.")

        # passed canary_percent is ignored when creating or restarting a service or multiple versions are being deployed.
        is_new_or_restarting = (
            existing_service is None
            or existing_service.current_state == ServiceEventCurrentState.TERMINATED
        )
        if canary_percent is not None and (
            is_new_or_restarting or versions is not None
        ):
            canary_percent = None
            self.logger.warning(
                "canary_percent is ignored when creating or restarting a service or multiple versions are being deployed."
            )

        rollout_strategy = self._get_rollout_strategy(
            existing_service, versions, in_place, configs  # type: ignore
        )

        service: DecoratedProductionServiceV2APIModel
        # TODO(doyoung): Support multi version deployment for in_place update.
        if rollout_strategy == RolloutStrategy.IN_PLACE:
            if versions is not None:
                raise ValueError(
                    "versions cannot be specified when doing an in_place update."
                )
            assert len(configs) == 1
            model = self._build_apply_service_model_for_in_place_update(
                name,
                configs[0],
                canary_percent=canary_percent,
                max_surge_percent=max_surge_percent,
                existing_service=existing_service,
            )
            service = self.client.rollout_service(model)
        elif rollout_strategy == RolloutStrategy.MULTI_VERSION:
            model = self._build_apply_service_model_for_multi_version(
                name,
                cloud,
                project,
                configs,
                existing_service=existing_service,
                versions=versions,  # type: ignore
            )
            service = self.client.rollout_service_multi_version(model)
        else:
            # rolling out a single version in new service without specifying versions.
            assert len(configs) == 1
            model = self._build_apply_service_model_for_rollout(
                name,
                configs[0],
                RolloutStrategy.ROLLOUT,
                canary_percent=canary_percent,
                max_surge_percent=max_surge_percent,
                existing_service=existing_service,
            )
            service = self.client.rollout_service(model)

        self._log_deployed_service_info(service, canary_percent=canary_percent)
        return service.id

    def _resolve_to_service_model(
        self,
        *,
        name: Optional[str] = None,
        cloud: Optional[str] = None,
        project: Optional[str] = None,
        include_archived: bool = False,
    ) -> DecoratedProductionServiceV2APIModel:
        if name is None:
            name = self._get_default_name()

        model = self.client.get_service(
            name=name, cloud=cloud, project=project, include_archived=include_archived
        )
        if model is None:
            raise RuntimeError(f"Service with name '{name}' was not found.")

        return model

    def rollback(
        self,
        name: Optional[str] = None,
        *,
        cloud: Optional[str] = None,
        project: Optional[str] = None,
        max_surge_percent: Optional[int] = None,
    ) -> str:
        model = self._resolve_to_service_model(name=name, cloud=cloud, project=project)

        self.client.rollback_service(model.id, max_surge_percent=max_surge_percent)

        return model.id

    def terminate(
        self,
        id: Optional[str] = None,  # noqa: A002
        name: Optional[str] = None,
        *,
        cloud: Optional[str] = None,
        project: Optional[str] = None,
    ) -> str:
        if id is not None:
            self.client.terminate_service(id)
            return id

        model = self._resolve_to_service_model(name=name, cloud=cloud, project=project)

        self.client.terminate_service(model.id)

        return model.id

    def archive(
        self,
        id: Optional[str] = None,  # noqa: A002
        name: Optional[str] = None,
        *,
        cloud: Optional[str] = None,
        project: Optional[str] = None,
    ) -> str:
        if id:
            self.client.archive_service(id)
            return id

        model = self._resolve_to_service_model(name=name, cloud=cloud, project=project)
        self.client.archive_service(model.id)

        return model.id

    def delete(
        self,
        id: Optional[str] = None,  # noqa: A002
        name: Optional[str] = None,
        *,
        cloud: Optional[str] = None,
        project: Optional[str] = None,
    ) -> str:
        if id:
            self.client.delete_service(id)
            return id

        model = self._resolve_to_service_model(
            name=name, cloud=cloud, project=project, include_archived=True
        )
        self.client.delete_service(model.id)

        return model.id

    def _get_user_facing_service_version_id(
        self, model: ProductionServiceV2VersionModel
    ) -> str:
        # NOTE(edoakes): the "version ID" exposed in the UI and tagged in the metrics is
        # not actually the DB ID, but a truncated version of it. We should store this
        # in the DB to avoid breakages, but for now I'm copying the existing logic.
        return model.id[-SERVICE_VERSION_ID_TRUNCATED_LEN:]

    async def _service_version_model_to_status_async(
        self,
        model: ProductionServiceV2VersionModel,
        *,
        service_name: str,
        project_id: str,
        query_auth_token_enabled: bool,
    ) -> ServiceVersionStatus:
        image_uri, image_build, project, compute_config = await asyncio.gather(
            asyncio.to_thread(
                self._image_sdk.get_image_uri_from_build_id, model.build_id
            ),
            asyncio.to_thread(self._image_sdk.get_image_build, model.build_id),
            asyncio.to_thread(self.client.get_project, project_id),
            asyncio.to_thread(
                self.get_user_facing_compute_config, model.compute_config_id
            ),
        )

        if image_uri is None:
            raise RuntimeError(f"Failed to get image URI for ID {model.build_id}.")
        if image_build is None:
            raise RuntimeError(f"Failed to get image build for ID {model.build_id}.")

        ray_gcs_external_storage_config = None
        if model.ray_gcs_external_storage_config is not None:
            ray_gcs_external_storage_config = RayGCSExternalStorageConfig(
                enabled=model.ray_gcs_external_storage_config.enable,
                address=model.ray_gcs_external_storage_config.address,
                certificate_path=model.ray_gcs_external_storage_config.redis_certificate_path,
            )

        tracing_config = None
        if model.tracing_config is not None:
            tracing_config = TracingConfig(
                enabled=model.tracing_config.enabled,
                exporter_import_path=model.tracing_config.exporter_import_path,
                sampling_ratio=model.tracing_config.sampling_ratio,
            )

        version_name = model.version or self._get_user_facing_service_version_id(model)

        return ServiceVersionStatus(
            id=self._get_user_facing_service_version_id(model),
            name=version_name,
            created_at=model.created_at,
            state=model.current_state,
            # NOTE(edoakes): there is also a "current_weight" field but it does not match the UI.
            weight=model.weight,
            config=ServiceConfig(
                name=service_name,
                version_name=version_name,
                applications=model.ray_serve_config["applications"],
                image_uri=str(image_uri),
                compute_config=compute_config,
                registry_login_secret=image_build.registry_login_secret,
                query_auth_token_enabled=query_auth_token_enabled,
                http_options=model.ray_serve_config.get("http_options", None),
                grpc_options=model.ray_serve_config.get("grpc_options", None),
                logging_config=model.ray_serve_config.get("logging_config", None),
                ray_gcs_external_storage_config=ray_gcs_external_storage_config,
                cloud=compute_config.cloud
                if compute_config and isinstance(compute_config, ComputeConfig)
                else None,
                project=project.name
                if project is not None and project.name != "default"
                else None,
                tracing_config=tracing_config,
            ),
        )

    def _model_state_to_state(
        self, model_state: ServiceEventCurrentState,
    ) -> ServiceState:
        # If we add a new state to the backend, old clients may not recognize it.
        # Rather than erroring out and causing old code to crash, return UNKNOWN.
        state: ServiceState = ServiceState(ServiceState.UNKNOWN)
        try:
            state = ServiceState(model_state)
        except ValueError:
            self.logger.warning(
                f"Got unrecognized state: '{model_state}'. "
                "You likely need to update the 'anyscale' package. "
                "If you still see this message after upgrading, contact Anyscale support."
            )

        return state

    async def _service_model_to_status_async(
        self, model: DecoratedProductionServiceV2APIModel
    ) -> ServiceStatus:
        # TODO(edoakes): this is currently only exposed at the service level in the API,
        # which means that the per-version `query_auth_token_enabled` field will lie if
        # it's changed.
        query_auth_token_enabled = model.auth_token is not None
        all_versions = None
        primary_version = None
        canary_version = None
        project_name = None

        if model.is_multi_version:
            # For multi-version services, format all versions
            version_tasks = [
                asyncio.create_task(
                    self._service_version_model_to_status_async(
                        version,
                        service_name=model.name,
                        project_id=model.project_id,
                        query_auth_token_enabled=query_auth_token_enabled,
                    )
                )
                for version in model.versions
            ]
            all_versions = await asyncio.gather(*version_tasks)

            if (
                all_versions
                and all_versions[0]
                and isinstance(all_versions[0].config, ServiceConfig)
            ):
                project_name = all_versions[0].config.project

        else:
            primary_version_task = None
            if model.primary_version is not None:
                primary_version_task = asyncio.create_task(
                    self._service_version_model_to_status_async(
                        model.primary_version,
                        service_name=model.name,
                        project_id=model.project_id,
                        query_auth_token_enabled=query_auth_token_enabled,
                    )
                )

            canary_version_task = None
            if model.canary_version is not None:
                canary_version_task = asyncio.create_task(
                    self._service_version_model_to_status_async(
                        model.canary_version,
                        service_name=model.name,
                        project_id=model.project_id,
                        query_auth_token_enabled=query_auth_token_enabled,
                    )
                )

            primary_version = (
                await primary_version_task if primary_version_task else None
            )
            canary_version = await canary_version_task if canary_version_task else None

            if primary_version and isinstance(primary_version.config, ServiceConfig):
                project_name = primary_version.config.project

        return ServiceStatus(
            id=model.id,
            name=model.name,
            creator=model.creator.email if model.creator else None,
            state=self._model_state_to_state(model.current_state),
            query_url=model.base_url,
            query_auth_token=model.auth_token,
            primary_version=primary_version,
            canary_version=canary_version,
            project=project_name,
            versions=all_versions,
        )

    def status(
        self, name: str, *, cloud: Optional[str] = None, project: Optional[str] = None
    ) -> ServiceStatus:
        model = self._resolve_to_service_model(name=name, cloud=cloud, project=project)
        return asyncio.run(self._service_model_to_status_async(model))

    def list(  # noqa: PLR0912, PLR0913
        self,
        *,
        # Single-item lookup
        service_id: Optional[str] = None,
        # Filters
        name: Optional[str] = None,
        state_filter: Optional[Union[List[ServiceState], List[str]]] = None,
        creator_id: Optional[str] = None,
        cloud: Optional[str] = None,
        project: Optional[str] = None,
        include_archived: bool = False,
        tags_filter: Optional[Dict[str, List[str]]] = None,
        # Paging
        max_items: Optional[int] = None,  # Controls total items yielded by iterator
        page_size: Optional[int] = None,  # Controls items fetched per API call
        # Sorting
        sort_field: Optional[Union[str, ServiceSortField]] = None,
        sort_order: Optional[Union[str, ServiceSortOrder]] = None,
    ) -> ResultIterator[ServiceStatus]:
        if page_size is not None and (page_size <= 0 or page_size > MAX_PAGE_SIZE):
            raise ValueError(
                f"page_size must be between 1 and {MAX_PAGE_SIZE}, inclusive."
            )

        if service_id is not None:
            raw = self.client.get_service_by_id(service_id)

            def _fetch_single_page(
                _token: Optional[str],
            ) -> DecoratedlistserviceapimodelListResponse:
                # Only return data on the first call (token=None), simulate single-item page
                if _token is None and raw is not None:
                    results = [raw]
                    metadata = ListResponseMetadata(total=1, next_paging_token=None)
                else:
                    results = []
                    metadata = ListResponseMetadata(total=0, next_paging_token=None)

                return DecoratedlistserviceapimodelListResponse(
                    results=results, metadata=metadata,
                )

            return ResultIterator(
                page_token=None,
                max_items=1,
                fetch_page=_fetch_single_page,
                async_parse_fn=self._service_model_to_status_async,
                parse_fn=None,
            )

        normalised_states = _normalize_state_filter(state_filter)

        # Convert tags dict into backend tag_filter list[str]
        backend_tag_filter: Optional[List[str]] = None
        if tags_filter:
            flattened: List[str] = []
            for key, values in tags_filter.items():
                if not values:
                    continue
                for value in values:
                    if key and value:
                        flattened.append(f"{key}:{value}")
            backend_tag_filter = flattened if len(flattened) > 0 else None

        def _fetch_page(
            token: Optional[str],
        ) -> DecoratedlistserviceapimodelListResponse:
            return self.client.list_services(
                name=name,
                state_filter=normalised_states,
                creator_id=creator_id,
                cloud=cloud,
                project=project,
                include_archived=include_archived,
                tag_filter=backend_tag_filter,
                count=page_size,
                paging_token=token,
                sort_field=sort_field,
                sort_order=sort_order,
            )

        return ResultIterator(
            page_token=None,
            max_items=max_items,
            fetch_page=_fetch_page,
            async_parse_fn=self._service_model_to_status_async,
            parse_fn=None,
        )

    def wait(
        self,
        name: str,
        *,
        cloud: Optional[str] = None,
        project: Optional[str] = None,
        state: ServiceState,
        timeout_s: float,
        interval_s: float,
    ):
        if not isinstance(timeout_s, (int, float)):
            raise TypeError("timeout_s must be a float")
        if timeout_s <= 0:
            raise ValueError("timeout_s must be >= 0")

        if not isinstance(interval_s, (int, float)):
            raise TypeError("interval_s must be a float")
        if interval_s <= 0:
            raise ValueError("interval_s must be >= 0")

        def _get_curr_state() -> ServiceState:
            model = self._resolve_to_service_model(
                name=name, cloud=cloud, project=project
            )

            return self._model_state_to_state(model.current_state)

        curr_state = _get_curr_state()
        self.logger.info(
            f"Waiting for service '{name}' to reach target state {state}, currently in state: {curr_state}"
        )
        for _ in self.timer.poll(timeout_s=timeout_s, interval_s=interval_s):
            new_state = _get_curr_state()
            if new_state != curr_state:
                self.logger.info(
                    f"Service '{name}' transitioned from {curr_state} to {new_state}"
                )
                curr_state = new_state

            if curr_state == state:
                self.logger.info(f"Service '{name}' reached target state, exiting")
                break
        else:
            raise TimeoutError(
                f"Service '{name}' did not reach target state {state} within {timeout_s}s. Last seen state: {curr_state}."
            )

    def controller_logs(
        self,
        name: str,
        *,
        cloud: Optional[str] = None,
        project: Optional[str] = None,
        canary: bool = False,
        mode: Union[str, ServiceLogMode] = ServiceLogMode.TAIL,
        max_lines: int = 1000,
    ) -> str:
        model = self._resolve_to_service_model(name=name, cloud=cloud, project=project)
        head = mode == ServiceLogMode.HEAD

        if canary:
            if model.canary_version is None:
                raise ValueError(
                    f"Service '{name}' is not currently rolling out. There is no canary version."
                )

            return self.client.controller_logs_for_service_version(
                model.canary_version, head, max_lines
            )

        if model.primary_version is None:
            raise ValueError(
                f"Service '{name}' is not ready yet. Please try again later..."
            )

        return self.client.controller_logs_for_service_version(
            model.primary_version, head, max_lines
        )

    def add_tags(
        self,
        *,
        id: Optional[str] = None,  # noqa: A002
        name: Optional[str] = None,
        cloud: Optional[str] = None,
        project: Optional[str] = None,
        tags: Dict[str, str],
    ) -> None:
        if not tags:
            raise ValueError("At least one tag must be provided.")

        if id is not None:
            resource_id = id
        else:
            if name is None and not self.client.inside_workspace():
                raise ValueError(
                    "Either 'id' or 'name' must be provided when running outside of a workspace."
                )
            model = self._resolve_to_service_model(
                name=name, cloud=cloud, project=project
            )
            resource_id = model.id

        self.client.upsert_resource_tags(
            ResourceTagResourceType.SERVICE, resource_id, tags
        )

    def remove_tags(
        self,
        *,
        id: Optional[str] = None,  # noqa: A002
        name: Optional[str] = None,
        cloud: Optional[str] = None,
        project: Optional[str] = None,
        keys: List[str],
    ) -> None:
        if not keys:
            raise ValueError("At least one tag key must be provided.")

        if id is not None:
            resource_id = id
        else:
            if name is None and not self.client.inside_workspace():
                raise ValueError(
                    "Either 'id' or 'name' must be provided when running outside of a workspace."
                )
            model = self._resolve_to_service_model(
                name=name, cloud=cloud, project=project
            )
            resource_id = model.id

        self.client.delete_resource_tags(
            ResourceTagResourceType.SERVICE, resource_id, keys
        )

    def list_tags(
        self,
        *,
        id: Optional[str] = None,  # noqa: A002
        name: Optional[str] = None,
        cloud: Optional[str] = None,
        project: Optional[str] = None,
    ) -> Dict[str, str]:
        """List tags for a service as a key/value mapping."""
        if id is not None:
            resource_id = id
        else:
            model = self._resolve_to_service_model(
                name=name, cloud=cloud, project=project
            )
            resource_id = model.id
        records = self.client.list_resource_tags(
            ResourceTagResourceType.SERVICE, resource_id
        )
        return {r.key: r.value for r in records if r and r.key is not None}


def _normalize_state_filter(
    states: Optional[Union[List[ServiceState], List[str]]],
) -> Optional[List[str]]:
    if states is None:
        return None

    normalized: List[str] = []
    for s in states:
        if isinstance(s, ServiceState):
            normalized.append(s.value)
        elif isinstance(s, str):
            normalized.append(s.upper())
        else:
            raise TypeError(
                "'state_filter' entries must be ServiceState or str, "
                f"got {type(s).__name__}"
            )
    return normalized



# AUTOGENERATED - modify shared_anyscale_util in root directory to make changes
import json
import os

import yaml


# TODO(nikita): Remove CLUSTER_YAML_TEMPLATE and CLUSTER_CONFIG_TEMPLATE_STR once we remove cluster config
# from projects model.
with open(os.path.join(os.path.dirname(__file__), "default_anyscale_aws.yaml")) as f:
    CLUSTER_YAML_TEMPLATE = f.read()

CLUSTER_CONFIG_TEMPLATE_STR = json.dumps(
    yaml.load(CLUSTER_YAML_TEMPLATE, Loader=yaml.SafeLoader)
)



import itertools
from json import dumps as json_dumps
from typing import Any, Callable, Dict, Iterator, List, Optional

import click
from rich.console import Console
from rich.table import Table

from anyscale.util import AnyscaleJSONEncoder, validate_non_negative_arg


MAX_PAGE_SIZE = 50
NON_INTERACTIVE_DEFAULT_MAX_ITEMS = 10


def validate_page_size(ctx, param, value):
    """Click callback to validate page size argument."""
    value = validate_non_negative_arg(ctx, param, value)
    if value is not None and value > MAX_PAGE_SIZE:
        raise click.BadParameter(f"must be less than or equal to {MAX_PAGE_SIZE}.")
    return value


def create_table(
    columns: List[tuple[str, Optional[str], bool]], is_first: bool = True
) -> Table:
    """Create a Rich table with specified columns.

    Args:
        columns: List of (column_name, style, no_wrap) tuples.
                 E.g., [("ID", "cyan", True), ("Name", None, False)]
        is_first: Show headers (True for first page, False for subsequent pages).

    Returns:
        Configured Rich Table ready for adding rows.

    Example:
        >>> columns = [("ID", "cyan", True), ("Name", None, False)]
        >>> table = create_table(columns, is_first=True)
        >>> table.add_row("123", "My Job")
    """
    table = Table(show_header=is_first, header_style="bold")
    for name, style, no_wrap in columns:
        table.add_column(name, style=style, no_wrap=no_wrap)
    return table


def _paginate(iterator: Iterator[Any], page_size: Optional[int]) -> Iterator[List[Any]]:
    if page_size is None:
        yield list(iterator)
    else:
        while True:
            page = list(itertools.islice(iterator, page_size))
            if not page:
                return
            yield page


def _render_page(
    page: List[Any],
    item_formatter: Callable[[Any], Dict[str, Any]],
    table_creator: Callable[[bool], Table],
    json_output: bool,
    is_first: bool,
    page_num: int,
    console: Console,
) -> int:
    """Render a single page of items."""
    if page_num > 1:  # Only show page number for pages after first
        console.print(f"[dim]Page {page_num}[/dim]")

    rows = [item_formatter(item) for item in page]
    if json_output:
        json_str = json_dumps(rows, indent=2, cls=AnyscaleJSONEncoder)
        console.print_json(json=json_str)
    else:
        tbl = table_creator(is_first)
        for row in rows:
            tbl.add_row(*row.values())
        console.print(tbl)

    return len(page)


def _should_continue_pagination(
    page_size: int, current_page_size: int, console: Console
) -> bool:
    """Prompt user to continue pagination if needed."""
    if current_page_size < page_size:
        return False  # Last page, no need to prompt

    console.print()
    console.print(
        "[dim]Press [bold]Enter[/bold] to continue, [bold]q[/bold] to quit…[/]"
    )
    return input("> ").strip().lower() != "q"


def display_list(  # noqa: PLR0913, PLR0912
    iterator: Iterator[Any],
    item_formatter: Callable[[Any], Dict[str, Any]],
    table_creator: Callable[[bool], Table],
    json_output: bool,
    page_size: int,
    interactive: bool,
    max_items: Optional[int],
    console: Console,
) -> int:
    """Displays a list of items from an iterator, handling pagination and output format.

    Args:
        iterator: The iterator yielding items to display.
        item_formatter: A callable that takes an item and returns a dictionary
            representing the row data (for table) or the JSON object.
        table_creator: A callable that takes a boolean (is_first_page) and
            returns a rich.Table instance. Used only if json_output is False.
        json_output: If True, output items as a JSON list. Otherwise, display
            them in a table created by table_creator.
        page_size: The number of items to display per page in interactive mode.
        interactive: If True, enables interactive pagination. If False, displays
            up to max_items (or all items if max_items is None) without prompting.
        max_items: The maximum total number of items to display when interactive
            is False. If None, all items are displayed.
        console: The rich.Console object to use for output.

    Returns:
        The total number of items displayed.
    """
    total_count = 0
    pages = _paginate(iterator, page_size if interactive else max_items)

    # Start interactive session if needed
    if interactive:
        try:
            from anyscale.telemetry import (  # noqa: PLC0415 - codex_reason("gpt5.2", "lazy import to avoid telemetry startup unless interactive")
                start_interactive_session,
            )

            start_interactive_session()
        except Exception:  # noqa: BLE001
            pass

    # Fetch and render first page
    with console.status("Retrieving items…", spinner="dots"):
        try:
            first_page = next(pages)
        except StopIteration:
            first_page = []

    if first_page:
        total_count += _render_page(
            first_page, item_formatter, table_creator, json_output, True, 1, console
        )

    # For interactive commands, mark when command logic completes
    if interactive:
        try:
            from anyscale.telemetry import (  # noqa: PLC0415 - codex_reason("gpt5.2", "lazy import to avoid telemetry startup unless interactive")
                mark_command_complete,
            )

            mark_command_complete()
        except Exception:  # noqa: BLE001
            pass

    # Non-interactive: stop after first page
    if not interactive:
        return total_count

    # Interactive: check if user wants to continue
    if not _should_continue_pagination(page_size, len(first_page), console):
        return total_count

    # Render remaining pages with correct telemetry timing
    page_num = 2
    while True:
        # Start page fetch timing and generate new trace ID BEFORE fetching
        try:
            from anyscale.telemetry import (  # noqa: PLC0415 - codex_reason("gpt5.2", "lazy import to avoid telemetry startup unless interactive")
                mark_page_fetch_start,
            )

            mark_page_fetch_start(page_num)
        except Exception:  # noqa: BLE001
            pass

        # Now fetch the page (with the new trace ID)
        try:
            page = next(pages)
        except StopIteration:
            break

        # Render the page
        total_count += _render_page(
            page, item_formatter, table_creator, json_output, False, page_num, console
        )

        # Complete page fetch telemetry
        try:
            from anyscale.telemetry import (  # noqa: PLC0415 - codex_reason("gpt5.2", "lazy import to avoid telemetry startup unless interactive")
                mark_page_fetch_complete,
            )

            mark_page_fetch_complete(page_num)
        except Exception:  # noqa: BLE001
            pass

        # Check if user wants to continue or if this was the last page
        if not _should_continue_pagination(page_size, len(page), console):
            break

        page_num += 1

    return total_count



from anyscale._private.anyscale_client.anyscale_client import (
    AnyscaleClient as AnyscaleClient,
)
from anyscale._private.anyscale_client.common import (
    AnyscaleClientInterface as AnyscaleClientInterface,
    DEFAULT_PYTHON_VERSION as DEFAULT_PYTHON_VERSION,
    DEFAULT_RAY_VERSION as DEFAULT_RAY_VERSION,
    WORKSPACE_CLUSTER_NAME_PREFIX as WORKSPACE_CLUSTER_NAME_PREFIX,
)
from anyscale._private.anyscale_client.fake_anyscale_client import (
    FakeAnyscaleClient as FakeAnyscaleClient,
)
